#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This code implements the SIFT algorithm
# The SIFT algorithm belongs to the University of British Columbia. It is
# protected by patent US6711293. If you are on a country where this pattent
# applies (like the USA), please check if you are allowed to use it. The
# University of British Columbia does not require a license for its use for
# non-commercial research applications.
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#
"""

Contains a class for creating a plan, allocating arrays, compiling kernels and other things like that...
to calculate SIFT keypoints and descriptors.

This algorithm is patented: U.S. Patent 6,711,293: 
"Method and apparatus for identifying scale invariant features in an image and use of same for locating an object in an image", 
David Lowe's patent for the SIFT algorithm, March 23, 2004
"""

from __future__ import division, print_function, with_statement

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-07-19"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import time, math, os, logging, threading, sys
# import sys
import gc
import numpy
from .param import par
from .opencl import ocl, pyopencl

from .utils import calc_size, kernel_size  # , sizeof
logger = logging.getLogger("sift.plan")

class SiftPlan(object):
    """
    This class implements a way to calculate SIFT keypoints.
    
    
    How to calculate a set of SIFT keypoint on an image:

    siftp = sift.SiftPlan(img.shape,img.dtype,devicetype="GPU")
    kp = siftp.keypoints(img)

    kp is a nx132 array. the second dimension is composed of x,y, scale and angle as well as 128 floats describing the keypoint

    This SIFT algorithm is patented: U.S. Patent 6,711,293: 
    "Method and apparatus for identifying scale invariant features in an image and use of same for locating an object in an image", 
    """
    kernels = {"convolution":1024,  # key: name value max local workgroup size
               "preprocess": 1024,
               "algebra": 1024,
               "image":1024,
               "gaussian":1024,
               "reductions":1024,
               "orientation_cpu":1,
               "orientation_gpu":128,
               "keypoints_gpu1":(8, 4, 4),
               "keypoints_gpu2":(8, 8, 8),
               "keypoints_cpu":1,
               "memset":128, }
#               "keypoints":128}
    converter = {numpy.dtype(numpy.uint8):"u8_to_float",
                 numpy.dtype(numpy.uint16):"u16_to_float",
                 numpy.dtype(numpy.uint32):"u32_to_float",
                 numpy.dtype(numpy.int32):"s32_to_float",
                 numpy.dtype(numpy.int64):"s64_to_float",
#                    numpy.float64:"double_to_float",
                      }
    sigmaRatio = 2.0 ** (1.0 / par.Scales)
    PIX_PER_KP = 10  # pre_allocate buffers for keypoints
    dtype_kp = numpy.dtype([('x', numpy.float32),
                                ('y', numpy.float32),
                                ('scale', numpy.float32),
                                ('angle', numpy.float32),
                                ('desc', (numpy.uint8, 128))
                                ])

    def __init__(self, shape=None, dtype=None, devicetype="CPU", template=None,
                 profile=False, device=None, PIX_PER_KP=None,
                 max_workgroup_size=None, context=None, init_sigma=None):
        """
        Contructor of the class

        @param shape: shape of the input image
        @param dtype: data type of the input image
        @param devicetype: can be 'CPU' or 'GPU'
        @param template: extract shape and dtype from an image
        @param profile: collect timing info
        @param device: 2-tuple of integers
        @param PIX_PER_KP: number of keypoint pre-allocated: 1 for 10 pixel
        @param max_workgroup_size: set to 1 under macosX on CPU
        @param context: provide an external context
        """
        if init_sigma is None:
            init_sigma = par.InitSigma
        # no test on the values, just make sure it is a float
        self._init_sigma = float(init_sigma)
        self.buffers = {}
        self.programs = {}
        if template is not None:
            self.shape = template.shape
            self.dtype = template.dtype
        else:
            self.shape = shape
            self.dtype = numpy.dtype(dtype)
        if len(self.shape) == 3:
            self.RGB = True
            self.shape = self.shape[:2]
        elif len(self.shape) == 2:
            self.RGB = False
        else:
            raise RuntimeError("Unable to process image of shape %s" % (tuple(self.shape,)))
        if PIX_PER_KP :
            self.PIX_PER_KP = int(PIX_PER_KP)
        self.profile = bool(profile)
        if max_workgroup_size:
            self.max_workgroup_size = int(max_workgroup_size)
            self.kernels = {}
            for k, v in self.__class__.kernels.items():
                if isinstance(v, int):
                    self.kernels[k] = min(v, self.max_workgroup_size)
#                else:
#                    self.kernels[k] = tuple([1 for i in v])
        else:
            self.max_workgroup_size = None
        self.events = []
        self._sem = threading.Semaphore()
        self.scales = []  # in XY order
        self.procsize = []  # same as  procsize but with dimension in (X,Y) not (slow, fast)
        self.wgsize = []
        self.kpsize = None
        self.memory = None
        self.octave_max = None
        self.red_size = None
        self._calc_scales()
        self._calc_memory()
        self.LOW_END = 0
        if context:
            self.ctx = context
            device_name = self.ctx.devices[0].name.strip()
            platform_name = self.ctx.devices[0].platform.name.strip()
            platform = ocl.get_platform(platform_name)
            device = platform.get_device(device_name)
            self.device = platform.id, device.id
        else:
            if device is None:
                self.device = ocl.select_device(type=devicetype, memory=self.memory, best=True)
                if device is None:
                    self.device = ocl.select_device(memory=self.memory, best=True)
                    logger.warning('Unable to find suitable device, selecting device: %s,%s' % self.device)
            else:
                self.device = device
            self.ctx = pyopencl.Context(devices=[pyopencl.get_platforms()[self.device[0]].get_devices()[self.device[1]]])

        if profile:
            self.queue = pyopencl.CommandQueue(self.ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
        else:
            self.queue = pyopencl.CommandQueue(self.ctx)
        self._calc_workgroups()
        self._compile_kernels()
        self._allocate_buffers()
        self.debug = []
        self.cnt = numpy.empty(1, dtype=numpy.int32)
        self.devicetype = ocl.platforms[self.device[0]].devices[self.device[1]].type
        if (self.devicetype == "CPU"):
            self.USE_CPU = True
            if sys.platform == "darwin":
                logger.warning("MacOSX computer working on CPU: limiting workgroup size to 1 !")
                self.max_workgroup_size = 1
                self.kernels = {}
                for k, v in self.__class__.kernels.items():
                    if isinstance(v, int):
                        self.kernels[k] = 1
                    else:
                        self.kernels[k] = tuple([1 for i in v])
        else:
            self.USE_CPU = False
            if "HD Graphics" in ocl.platforms[self.device[0]].devices[self.device[1]].name:
                self.LOW_END=2
            
    def __del__(self):
        """
        Destructor: release all buffers
        """
        self._free_kernels()
        self._free_buffers()
        self.queue = None
        self.ctx = None
        gc.collect()

    def _calc_scales(self):
        """
        Nota scales are in XY order
        """
        shape = self.shape[-1::-1]
        self.scales = [tuple(numpy.int32(i) for i in shape)]
        min_size = 2 * par.BorderDist + 2
        while min(shape) > min_size:
            shape = tuple(numpy.int32(i // 2) for i in shape)
            self.scales.append(shape)
        self.scales.pop()
        self.octave_max = len(self.scales)

    def _calc_memory(self):
        """
        Estimates the memory footprint of all buffer to ensure it fits on the device
        """
        # Just the context + kernel takes about 75MB on the GPU
        self.memory = 75 * 2 ** 20
        size_of_float = numpy.dtype(numpy.float32).itemsize
        size_of_input = numpy.dtype(self.dtype).itemsize
        # raw images:
        size = self.shape[0] * self.shape[1]
        self.memory += size * size_of_input  # initial_image (no raw_float)
        if self.RGB:
            self.memory += 2 * size * (size_of_input)  # one of three was already counted
        nr_blur = par.Scales + 3  # 3 blurs and 2 tmp
        nr_dogs = par.Scales + 2
        self.memory += size * (nr_blur + nr_dogs) * size_of_float

        self.kpsize = int(self.shape[0] * self.shape[1] // self.PIX_PER_KP)  # Is the number of kp independant of the octave ? int64 causes problems with pyopencl
        self.memory += self.kpsize * size_of_float * 4 * 2  # those are array of float4 to register keypoints, we need two of them
        self.memory += self.kpsize * 128  # stores the descriptors: 128 unsigned chars
        self.memory += 4  # keypoint index Counter
        if self.max_workgroup_size:
            wg_float = min(self.max_workgroup_size, numpy.sqrt(self.shape[0] * self.shape[1]))
        else:
            wg_float = min(128, numpy.sqrt(self.shape[0] * self.shape[1]))
        self.red_size = 2 ** (int(math.ceil(math.log(wg_float, 2))))
        self.memory += 4 * 2 * self.red_size  # temporary storage for reduction

        ########################################################################
        # Calculate space for gaussian kernels
        ########################################################################
        curSigma = 1.0 if par.DoubleImSize else 0.5
        if self._init_sigma > curSigma:
            sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
            size = kernel_size(sigma, True)
            logger.debug("pre-Allocating %s float for init blur" % size)
            self.memory += size * size_of_float
        prevSigma = self._init_sigma
        for i in range(par.Scales + 2):
            increase = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            size = kernel_size(increase, True)
            logger.debug("pre-Allocating %s float for blur sigma: %s" % (size, increase))
            self.memory += size * size_of_float
            prevSigma *= self.sigmaRatio

    def _allocate_buffers(self):
        """
        All buffers are allocated here
        """
        shape = self.shape
        if self.dtype != numpy.float32:
            if self.RGB:
                rgbshape = self.shape[0], self.shape[1], 3
                self.buffers["raw"] = pyopencl.array.empty(self.queue, rgbshape, dtype=self.dtype)
            else:
                self.buffers["raw"] = pyopencl.array.empty(self.queue, shape, dtype=self.dtype)
        self.buffers[ "Kp_1" ] = pyopencl.array.empty(self.queue, (self.kpsize, 4), dtype=numpy.float32)
        self.buffers[ "Kp_2" ] = pyopencl.array.empty(self.queue, (self.kpsize, 4), dtype=numpy.float32)
        self.buffers[ "descr" ] = pyopencl.array.empty(self.queue, (self.kpsize, 128), dtype=numpy.uint8)
        self.buffers["cnt" ] = pyopencl.array.empty(self.queue, 1, dtype=numpy.int32)
        self.buffers["descriptors"] = pyopencl.array.empty(self.queue, (self.kpsize, 128), dtype=numpy.uint8)

        self.buffers["tmp"] = pyopencl.array.empty(self.queue, shape, dtype=numpy.float32)
        self.buffers["ori"] = pyopencl.array.empty(self.queue, shape, dtype=numpy.float32)
        for scale in range(par.Scales + 3):
            self.buffers[scale ] = pyopencl.array.empty(self.queue, shape, dtype=numpy.float32)
        self.buffers["DoGs" ] = pyopencl.array.empty(self.queue, (par.Scales + 2, shape[0], shape[1]), dtype=numpy.float32)
        wg_float = min(512.0, numpy.sqrt(self.shape[0] * self.shape[1]))
#        wg = 2 ** (int(math.ceil(math.log(wg_float, 2))))
        self.buffers["max_min"] = pyopencl.array.empty(self.queue, (self.red_size, 2), dtype=numpy.float32)  # temporary buffer for max/min reduction
        self.buffers["min"] = pyopencl.array.empty(self.queue, (1), dtype=numpy.float32)
        self.buffers["max"] = pyopencl.array.empty(self.queue, (1), dtype=numpy.float32)
        self.buffers["255"] = pyopencl.array.to_device(self.queue, numpy.array([255.0], dtype=numpy.float32))
        ########################################################################
        # Allocate space for gaussian kernels
        ########################################################################
        curSigma = 1.0 if par.DoubleImSize else 0.5
        if self._init_sigma > curSigma:
            sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
            self._init_gaussian(sigma)
        prevSigma = self._init_sigma

        for i in range(par.Scales + 2):
            increase = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            self._init_gaussian(increase)
            prevSigma *= self.sigmaRatio


    def _init_gaussian(self, sigma):
        """
        Create a buffer of the right size according to the width of the gaussian ...


        @param  sigma: width of the gaussian, the length of the function will be 8*sigma + 1

        Same calculation done on CPU
        x = numpy.arange(size) - (size - 1.0) / 2.0
        gaussian = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
        gaussian /= gaussian.sum(dtype=numpy.float32)
        """
        name = "gaussian_%s" % sigma
        size = kernel_size(sigma, True)
        wg_size = 2 ** int(math.ceil(math.log(size) / math.log(2)))

        logger.debug("Allocating %s float for blur sigma: %s" % (size, sigma))
        if self.max_workgroup_size and (wg_size > self.max_workgroup_size):  # compute on CPU
            x = numpy.arange(size) - (size - 1.0) / 2.0
            gaus = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
            gaus /= gaus.sum(dtype=numpy.float32)
            gaussian_gpu = pyopencl.array.to_device(self.queue, gaus)
        else:
            gaussian_gpu = pyopencl.array.empty(self.queue, size, dtype=numpy.float32)
            evt = self.programs["gaussian"].gaussian(self.queue, (wg_size,), (wg_size,),
                                                gaussian_gpu.data,  # __global     float     *data,
                                                numpy.float32(sigma),  # const        float     sigma,
                                                numpy.int32(size))  # const        int     SIZE
            if self.profile: self.events.append(("gaussian %s" % sigma, evt))
        self.buffers[name] = gaussian_gpu


    def _free_buffers(self):
        """
        free all memory allocated on the device
        """
        for buffer_name in self.buffers:
            if self.buffers[buffer_name] is not None:
                try:
                    del self.buffers[buffer_name]
                    self.buffers[buffer_name] = None
                except pyopencl.LogicError:
                    logger.error("Error while freeing buffer %s" % buffer_name)

    def _compile_kernels(self):
        """
        Call the OpenCL compiler
        """
        kernel_directory = os.path.dirname(os.path.abspath(__file__))
        if not os.path.exists(os.path.join(kernel_directory, "algebra" + ".cl")):
            while (".zip" in kernel_directory)  and (len(kernel_directory) > 4):
                kernel_directory = os.path.dirname(kernel_directory)
            kernel_directory = os.path.join(kernel_directory, "sift_kernels")
        for kernel in self.kernels:
            kernel_file = os.path.join(kernel_directory, kernel + ".cl")
            kernel_src = open(kernel_file).read()
            if self.max_workgroup_size:
                if "__len__" not in dir(self.kernels[kernel]):
                    wg_size = min(self.max_workgroup_size, self.kernels[kernel])
                else:
                    wg_size = self.max_workgroup_size
            else:
                if "__len__" not in dir(self.kernels[kernel]):
                    wg_size = self.kernels[kernel]
                else:
                    wg_size = 128
            try:
                program = pyopencl.Program(self.ctx, kernel_src).build('-D WORKGROUP_SIZE=%s' % wg_size)
            except pyopencl.MemoryError as error:
                raise MemoryError(error)
            except pyopencl.RuntimeError as error:
                if kernel == "keypoints_gpu2":
                    logger.warning("Failed compiling kernel '%s' with workgroup size %s: %s: use low_end alternative", kernel, wg_size, error)
                    self.LOW_END += 1
                elif kernel == "keypoints_gpu1":
                    logger.warning("Failed compiling kernel '%s' with workgroup size %s: %s: use CPU alternative", kernel, wg_size, error)
                    self.LOW_END += 1
                else:
                    logger.error("Failed compiling kernel '%s' with workgroup size %s: %s", kernel, wg_size, error)
                    raise error
            self.programs[kernel] = program

    def _free_kernels(self):
        """
        free all kernels
        """
        self.programs = {}

    def _calc_workgroups(self):
        """
        First try to guess the best workgroup size, then calculate all global worksize

        Nota:
        The workgroup size is limited by the device
        The workgroup size is limited to the 2**n below then image size (hence changes with octaves)
        The second dimension of the wg size should be large, the first small: i.e. (1,64)
        The processing size should be a multiple of  workgroup size.
        """
        device = self.ctx.devices[0]
#        max_work_group_size = device.max_work_group_size
        max_work_item_sizes = device.max_work_item_sizes
        # we recalculate the shapes ...
        shape = self.shape
        min_size = 2 * par.BorderDist + 2
        if self.max_workgroup_size:
            self.max_workgroup_size = min(self.max_workgroup_size, max_work_item_sizes[0])
        else:
            self.max_workgroup_size = max_work_item_sizes[0]
        while min(shape) > min_size:
            wg = (min(2 ** int(math.ceil(math.log(shape[-1], 2))), self.max_workgroup_size), 1)
            self.wgsize.append(wg)
            self.procsize.append(calc_size(shape[-1::-1], wg))
            shape = tuple(i // 2 for i in shape)


    def keypoints(self, image):
        """
        Calculates the keypoints of the image
        @param image: ndimage of 2D (or 3D if RGB)
        """
        self.reset_timer()
        with self._sem:
            total_size = 0
            keypoints = []
            descriptors = []
            assert image.shape[:2] == self.shape
            assert image.dtype == self.dtype
            t0 = time.time()

            if self.dtype == numpy.float32:
                if type(image) == pyopencl.array.Array:
                    evt = pyopencl.enqueue_copy(self.queue, self.buffers[0].data, image.data)
                else:
                    evt = pyopencl.enqueue_copy(self.queue, self.buffers[0].data, image)
                if self.profile:self.events.append(("copy H->D", evt))
            elif (len(image.shape) == 3) and (self.dtype == numpy.uint8) and (self.RGB):
                if type(image) == pyopencl.array.Array:
                    evt = pyopencl.enqueue_copy(self.queue, self.buffers["raw"].data, image.data)
                else:
                    evt = pyopencl.enqueue_copy(self.queue, self.buffers["raw"].data, image)
                if self.profile:self.events.append(("copy H->D", evt))
#                print self.procsize[0], self.wgsize[0]
                evt = self.programs["preprocess"].rgb_to_float(self.queue, self.procsize[0], self.wgsize[0],
                                                             self.buffers["raw"].data, self.buffers[0].data, *self.scales[0])
                if self.profile:self.events.append(("RGB -> float", evt))

            elif self.dtype in self.converter:
                program = self.programs["preprocess"].__getattr__(self.converter[self.dtype])
                evt = pyopencl.enqueue_copy(self.queue, self.buffers["raw"].data, image)
                if self.profile:self.events.append(("copy H->D", evt))
                evt = program(self.queue, self.procsize[0], self.wgsize[0],
                        self.buffers["raw"].data, self.buffers[0].data, *self.scales[0])
                if self.profile:self.events.append(("convert -> float", evt))
            else:
                raise RuntimeError("invalid input format error")

            k1 = self.programs["reductions"].max_min_global_stage1(self.queue, (self.red_size * self.red_size,), (self.red_size,),
                                                                   self.buffers[0].data,
                                                                   self.buffers["max_min"].data,
                                                                   numpy.uint32(self.shape[0] * self.shape[1]))
            k2 = self.programs["reductions"].max_min_global_stage2(self.queue, (self.red_size,), (self.red_size,),
                                                                   self.buffers["max_min"].data,
                                                                   self.buffers["max"].data,
                                                                   self.buffers["min"].data)
            if self.profile:
                self.events.append(("max_min_stage1", k1))
                self.events.append(("max_min_stage2", k2))
            evt = self.programs["preprocess"].normalizes(self.queue, self.procsize[0], self.wgsize[0],
                                                   self.buffers[0].data,
                                                   self.buffers["min"].data,
                                                   self.buffers["max"].data,
                                                   self.buffers["255"].data,
                                                   *self.scales[0])
            if self.profile:self.events.append(("normalize", evt))


            curSigma = 1.0 if par.DoubleImSize else 0.5
            octave = 0
            if self._init_sigma > curSigma:
                logger.debug("Bluring image to achieve std: %f", self._init_sigma)
                sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
                self._gaussian_convolution(self.buffers[0], self.buffers[0], sigma, 0)

            for octave in range(self.octave_max):
                kp, descriptor = self._one_octave(octave)
                logger.info("in octave %i found %i kp" % (octave, kp.shape[0]))

                if kp.shape[0] > 0:
                    keypoints.append(kp)
                    descriptors.append(descriptor)
                    total_size += kp.shape[0]

            ########################################################################
            # Merge keypoints in central memory
            ########################################################################
            output = numpy.recarray(shape=(total_size,), dtype=self.dtype_kp)
            last = 0
            for ds, desc in zip(keypoints, descriptors):
                l = ds.shape[0]
                if l > 0:
                    output[last:last + l].x = ds[:, 0]
                    output[last:last + l].y = ds[:, 1]
                    output[last:last + l].scale = ds[:, 2]
                    output[last:last + l].angle = ds[:, 3]
                    output[last:last + l].desc = desc
                    last += l
            logger.info("Execution time: %.3fms" % (1000 * (time.time() - t0)))
        return output

    def _gaussian_convolution(self, input_data, output_data, sigma, octave=0):
        """
        Calculate the gaussian convolution with precalculated kernels.

        @param input_data: pyopencl array with input
        @param output_data: pyopencl array with result
        @param sigma: width of the gaussian
        @param octave: related to the size on the input images

        * Uses a temporary buffer
        * Needs gaussian kernel to be available on device

        """
        temp_data = self.buffers["tmp"]
        gaussian = self.buffers["gaussian_%s" % sigma]
        k1 = self.programs["convolution"].horizontal_convolution(self.queue, self.procsize[octave], self.wgsize[octave],
                                input_data.data, temp_data.data, gaussian.data, numpy.int32(gaussian.size), *self.scales[octave])
        k2 = self.programs["convolution"].vertical_convolution(self.queue, self.procsize[octave], self.wgsize[octave],
                                temp_data.data, output_data.data, gaussian.data, numpy.int32(gaussian.size), *self.scales[octave])

        if self.profile:
            self.events += [("Blur sigma %s octave %s" % (sigma, octave), k1), ("Blur sigma %s octave %s" % (sigma, octave), k2)]

    def _one_octave(self, octave):
        """
        Does all scales within an octave

        @param octave: number of the octave
        """
        prevSigma = self._init_sigma
        logger.info("Calculating octave %i" % octave)
        wgsize = (128,)  # (max(self.wgsize[octave]),) #TODO: optimize
        kpsize32 = numpy.int32(self.kpsize)
        self._reset_keypoints()
        octsize = numpy.int32(2 ** octave)
        last_start = numpy.int32(0)
        for scale in range(par.Scales + 2):
            sigma = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            logger.info("Octave %i scale %s blur with sigma %s" % (octave, scale, sigma))

            ########################################################################
            # Calculate gaussian blur and DoG
            ########################################################################

            self._gaussian_convolution(self.buffers[scale], self.buffers[scale + 1], sigma, octave)
            prevSigma *= self.sigmaRatio
            evt = self.programs["algebra"].combine(self.queue, self.procsize[octave], self.wgsize[octave],
                                             self.buffers[scale + 1].data, numpy.float32(-1.0),
                                             self.buffers[scale].data, numpy.float32(+1.0),
                                             self.buffers["DoGs"].data, numpy.int32(scale),
                                             *self.scales[octave])
            if self.profile:self.events.append(("DoG %s %s" % (octave, scale), evt))
        for scale in range(1, par.Scales + 1):
#                print("Before local_maxmin, cnt is %s %s %s" % (self.buffers["cnt"].get()[0], self.procsize[octave], self.wgsize[octave]))
            evt = self.programs["image"].local_maxmin(self.queue, self.procsize[octave], self.wgsize[octave],
                                            self.buffers["DoGs"].data,  # __global float* DOGS,
                                            self.buffers["Kp_1"].data,  # __global keypoint* output,
                                            numpy.int32(par.BorderDist),  # int border_dist,
                                            numpy.float32(par.PeakThresh),  # float peak_thresh,
                                            octsize,  # int octsize,
                                            numpy.float32(par.EdgeThresh1),  # float EdgeThresh0,
                                            numpy.float32(par.EdgeThresh),  # float EdgeThresh,
                                            self.buffers["cnt"].data,  # __global int* counter,
                                            kpsize32,  # int nb_keypoints,
                                            numpy.int32(scale),  # int scale,
                                            *self.scales[octave])  # int width, int height)
            if self.profile:self.events.append(("local_maxmin %s %s" % (octave, scale), evt))
            procsize = calc_size((self.kpsize,), wgsize)
            cp_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.buffers["cnt"].data)
            # TODO: modify interp_keypoint so that it reads end_keypoint from GPU memory
            evt = self.programs["image"].interp_keypoint(self.queue, procsize, wgsize,
                                          self.buffers["DoGs"].data,  # __global float* DOGS,
                                          self.buffers["Kp_1"].data,  # __global keypoint* keypoints,
                                          last_start,  # int start_keypoint,
                                          self.cnt[0],  # int end_keypoint,
                                          numpy.float32(par.PeakThresh),  # float peak_thresh,
                                          numpy.float32(self._init_sigma),  # float InitSigma,
                                          *self.scales[octave])  # int width, int height)
            if self.profile:
                self.events += [("get cnt", cp_evt),
                                ("interp_keypoint %s %s" % (octave, scale), evt)
                                ]

            newcnt = self._compact(last_start)
            evt = self.programs["image"].compute_gradient_orientation(self.queue, self.procsize[octave], self.wgsize[octave],
                               self.buffers[scale].data,  # __global float* igray,
                               self.buffers["tmp"].data,  # __global float *grad,
                               self.buffers["ori"].data,  # __global float *ori,
                               *self.scales[octave])  # int width,int height
            if self.profile:self.events.append(("compute_gradient_orientation %s %s" % (octave, scale), evt))

#           Orientation assignement: 1D kernel, rather heavy kernel
            if newcnt and newcnt > last_start:  # launch kernel only if neededwgsize = (128,)

                if self.USE_CPU:
                    file_to_use = "orientation_cpu"
#                    logger.info("Computing orientation with CPU-optimized kernels")
                else:
                    file_to_use = "orientation_gpu"

                wgsize2 = self.kernels[file_to_use],
                procsize = int(newcnt * wgsize2[0]),
                evt = self.programs[file_to_use].orientation_assignment(self.queue, procsize, wgsize2,
                                      self.buffers["Kp_1"].data,  # __global keypoint* keypoints,
                                      self.buffers["tmp"].data,  # __global float* grad,
                                      self.buffers["ori"].data,  # __global float* ori,
                                      self.buffers["cnt"].data,  # __global int* counter,
                                      octsize,  # int octsize,
                                      numpy.float32(par.OriSigma),  # float OriSigma, //WARNING: (1.5), it is not "InitSigma (=1.6)"
                                      kpsize32,  # int max of nb_keypoints,
                                      numpy.int32(last_start),  # int keypoints_start,
                                      newcnt,  # int keypoints_end,
                                      *self.scales[octave])  # int grad_width, int grad_height)
                # newcnt = self.buffers["cnt"].get()[0] #do not forget to update numbers of keypoints, modified above !
                evt_cp = pyopencl.enqueue_copy(self.queue, self.cnt, self.buffers["cnt"].data)
                newcnt = self.cnt[0]  # do not forget to update numbers of keypoints, modified above !

                if self.USE_CPU or self.LOW_END == 2:
                    file_to_use = "keypoints_cpu"
                    logger.info("Computing descriptors with CPU optimized kernels")
                    wgsize2 = self.kernels[file_to_use],
                    procsize2 = int(newcnt * wgsize2[0]),

                else:
                    if self.LOW_END == 1 :
                        file_to_use = "keypoints_gpu1"
                        logger.info("Computing descriptors with older-GPU optimized kernels")
                        wgsize2 = self.kernels[file_to_use]
                    else:
                        file_to_use = "keypoints_gpu2"
                        logger.info("Computing descriptors with newer-GPU optimized kernels")
                        wgsize2 = self.kernels[file_to_use]
                    procsize2 = int(newcnt * wgsize2[0]), wgsize2[1], wgsize2[2]
                try:
                    evt2 = self.programs[file_to_use].descriptor(self.queue, procsize2, wgsize2,
                                          self.buffers["Kp_1"].data,  # __global keypoint* keypoints,
                                          self.buffers["descriptors"].data,  # ___global unsigned char *descriptors
                                          self.buffers["tmp"].data,  # __global float* grad,
                                          self.buffers["ori"].data,  # __global float* ori,
                                          octsize,  # int octsize,
                                          numpy.int32(last_start),  # int keypoints_start,
                                          self.buffers["cnt"].data,  # int* keypoints_end,
                                          *self.scales[octave])  # int grad_width, int grad_height)
                except pyopencl.RuntimeError as error:
                    self.LOW_END += 1
                    logger.error("Descriptor failed with %s. Switching to lower_end mode" % error)
                    if self.USE_CPU or self.LOW_END == 2:
                        file_to_use = "keypoints_cpu"
                        logger.info("Computing descriptors with CPU optimized kernels")
                        wgsize2 = self.kernels[file_to_use],
                        procsize2 = int(newcnt * wgsize2[0]),

                    else:
                        if self.LOW_END == 1 :
                            file_to_use = "keypoints_gpu1"
                            logger.info("Computing descriptors with older-GPU optimized kernels")
                            wgsize2 = self.kernels[file_to_use]
                        else:
                            file_to_use = "keypoints_gpu2"
                            logger.info("Computing descriptors with newer-GPU optimized kernels")
                            wgsize2 = self.kernels[file_to_use]
                        procsize2 = int(newcnt * wgsize2[0]), wgsize2[1], wgsize2[2]

                    evt2 = self.programs[file_to_use].descriptor(self.queue, procsize2, wgsize2,
                                          self.buffers["Kp_1"].data,  # __global keypoint* keypoints,
                                          self.buffers["descriptors"].data,  # ___global unsigned char *descriptors
                                          self.buffers["tmp"].data,  # __global float* grad,
                                          self.buffers["ori"].data,  # __global float* ori,
                                          octsize,  # int octsize,
                                          numpy.int32(last_start),  # int keypoints_start,
                                          self.buffers["cnt"].data,  # int* keypoints_end,
                                          *self.scales[octave])  # int grad_width, int grad_height)
                if self.profile:
                    self.events += [("orientation_assignment %s %s" % (octave, scale), evt),
                                    ("copy cnt D->H", evt_cp),
                                    ("descriptors %s %s" % (octave, scale), evt2)]

            evt_cp = pyopencl.enqueue_copy(self.queue, self.cnt, self.buffers["cnt"].data)
            last_start = self.cnt[0]
            if self.profile:
                self.events.append(("copy cnt D->H", evt_cp))

        ########################################################################
        # Rescale all images to populate all octaves
        ########################################################################
        if octave < self.octave_max - 1:
            evt = self.programs["preprocess"].shrink(self.queue, self.procsize[octave + 1], self.wgsize[octave + 1],
                                                    self.buffers[par.Scales].data,
                                                    self.buffers[0].data,
                                                    numpy.int32(2), numpy.int32(2),
                                                    self.scales[octave][0], self.scales[octave][1],
                                                    *self.scales[octave + 1])
            if self.profile:
                self.events.append(("shrink %s->%s" % (self.scales[octave], self.scales[octave + 1]), evt))
        results = numpy.empty((last_start, 4), dtype=numpy.float32)
        descriptors = numpy.empty((last_start, 128), dtype=numpy.uint8)
        if last_start:
            evt = pyopencl.enqueue_copy(self.queue, results, self.buffers["Kp_1"].data)
            evt2 = pyopencl.enqueue_copy(self.queue, descriptors, self.buffers["descriptors"].data)
            if self.profile:
                self.events += [("copy D->H", evt),
                                ("copy D->H", evt2)]
        return results, descriptors

    def _compact(self, start=numpy.int32(0)):
        """
        Compact the vector of keypoints starting from start

        @param start: start compacting at this adress. Before just copy
        @type start: numpy.int32
        """
        wgsize = self.max_workgroup_size,  # (max(self.wgsize[0]),) #TODO: optimize
        kpsize32 = numpy.int32(self.kpsize)
        cp0_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.buffers["cnt"].data)
        kp_counter = self.cnt[0]
        procsize = calc_size((self.kpsize,), wgsize)

        if kp_counter > 0.9 * self.kpsize:
            logger.warning("Keypoint counter overflow risk: counted %s / %s" % (kp_counter, self.kpsize))
        logger.info("Compact %s -> %s / %s" % (start, kp_counter, self.kpsize))
        self.cnt[0] = start
        cp1_evt = pyopencl.enqueue_copy(self.queue, self.buffers["cnt"].data, self.cnt)
        evt = self.programs["algebra"].compact(self.queue, procsize, wgsize,
                        self.buffers["Kp_1"].data,  # __global keypoint* keypoints,
                        self.buffers["Kp_2"].data,  # __global keypoint* output,
                        self.buffers["cnt"].data,  # __global int* counter,
                        start,  # int start,
                        kp_counter)  # int nbkeypoints
        cp2_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.buffers["cnt"].data)
        # swap keypoints:
        self.buffers["Kp_1"], self.buffers["Kp_2"] = self.buffers["Kp_2"], self.buffers["Kp_1"]
        # memset buffer Kp_2
#        self.buffers["Kp_2"].fill(-1, self.queue)
        mem_evt = self.programs["memset"].memset_float(self.queue, calc_size((4 * self.kpsize,), wgsize), wgsize, self.buffers["Kp_2"].data, numpy.float32(-1), numpy.int32(4 * self.kpsize))
        if self.profile:
            self.events += [("copy cnt D->H", cp0_evt),
                            ("copy cnt H->D", cp1_evt),
                            ("compact", evt),
                            ("copy cnt D->H", cp2_evt),
                            ("memset 2", mem_evt)
                            ]
        return self.cnt[0]


    def _reset_keypoints(self):
        """
        Todo: implement directly in OpenCL instead of relying on pyOpenCL
        """
        wg_size = self.kernels["memset"],
        evt1 = self.programs["memset"].memset_float(self.queue, calc_size((4 * self.kpsize,), wg_size), wg_size, self.buffers["Kp_1"].data, numpy.float32(-1), numpy.int32(4 * self.kpsize))
#        evt2 = self.programs["memset"].memset_float(self.queue, calc_size((4 * self.kpsize,), wg_size), wg_size, self.buffers["Kp_2"].data, numpy.float32(-1), numpy.int32(4 * self.kpsize))
        evt3 = self.programs["memset"].memset_int(self.queue, (1,), (1,), self.buffers["cnt"].data, numpy.int32(0), numpy.int32(1))
        if self.profile:
            self.events += [("memset 1", evt1), ("memset cnt", evt3)]
#        self.buffers["Kp_1"].fill(-1, self.queue)
#        self.buffers["Kp_2"].fill(-1, self.queue)
#        self.buffers["cnt"].fill(0, self.queue)

    def count_kp(self, output):
        """
        Print the number of keypoint per octave
        """
        kpt = 0
        for octave, data in enumerate(output):
            if output.shape[0] > 0:
                ksum = (data[:, 1] != -1.0).sum()
                kpt += ksum
                print("octave %i kp count %i/%i size %s ratio:%s" % (octave, ksum, self.kpsize, self.scales[octave], 1000.0 * ksum / self.scales[octave][1] / self.scales[octave][0]))
        print("Found total %i guess %s pixels per keypoint" % (kpt, self.shape[0] * self.shape[1] / kpt))

    def debug_holes(self, label=""):
        print("%s %s" % (label, numpy.where(self.buffers["Kp_1"].get()[:, 1] == -1)[0]))

    def log_profile(self):
        """
        If we are in debugging mode, prints out all timing for every single OpenCL call
        """
        t = 0.0
        orient = 0.0
        descr = 0.0
        if self.profile:
            for e in self.events:
                if "__len__" in dir(e) and len(e) >= 2:
                    et = 1e-6 * (e[1].profile.end - e[1].profile.start)
                    print("%50s:\t%.3fms" % (e[0], et))
                    t += et
                    if "orient" in e[0]:
                        orient += et
                    if "descriptors" in e[0]:
                        descr += et

        print("_"*80)
        print("%50s:\t%.3fms" % ("Total execution time", t))
        print("%50s:\t%.3fms" % ("Total Orientation assignment", orient))
        print("%50s:\t%.3fms" % ("Total Descriptors", descr))

    def reset_timer(self):
        """
        Resets the profiling timers
        """
        with self._sem:
            self.events = []

if __name__ == "__main__":
    # Prepare debugging
    import scipy.misc
    lena = scipy.misc.lena()
    s = SiftPlan(template=lena)
    s.keypoints(lena)









#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Contains a class for creating a matching plan, allocating arrays, compiling kernels and other things like that
"""

from __future__ import division, print_function, with_statement

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-07-23"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import time, math, os, logging, sys, threading
import gc
import numpy
from .param import par
from .opencl import ocl, pyopencl
from .utils import calc_size, kernel_size, sizeof
logger = logging.getLogger("sift.match")
if pyopencl:
    from pyopencl import mem_flags as MF
else:
    logger.warning("No PyOpenCL, no sift")

class MatchPlan(object):
    """
    Plan to compare sets of SIFT keypoint

    siftp = sift.MatchPlan(devicetype="GPU")
    kp = siftp.match(kp1,kp2)

    kp is a nx132 array. the second dimension is composed of x,y, scale and angle as well as 128 floats describing the keypoint

    """
    kernels = {"matching_gpu":64,
               "matching_cpu":16,
               "memset":128, }

    dtype_kp = numpy.dtype([('x', numpy.float32),
                                ('y', numpy.float32),
                                ('scale', numpy.float32),
                                ('angle', numpy.float32),
                                ('desc', (numpy.uint8, 128))
                                ])

    def __init__(self, size=16384, devicetype="CPU", profile=False, device=None, max_workgroup_size=None, roi=None, context=None):
        """
        Constructor of the class:

        @param size: size of the input keypoint-list alocated on the GPU.
        @param devicetype: can be CPU or GPU
        @param profile: set to true to activate profiling information collection
        @param device: 2-tuple of integer, see clinfo
        @param max_workgroup_size: CPU on MacOS, limit to 1. None by default to use default ones (max=128). 
        @param roi: Region Of Interest: TODO
        @param context: Use an external context (discard devicetype and device options)
        """
        self.profile = bool(profile)
        if max_workgroup_size:
            self.max_workgroup_size = int(max_workgroup_size)
            self.kernels = {}
            for k, v in self.__class__.kernels.items():
                self.kernels[k] = min(v, self.max_workgroup_size)
        else:
            self.max_workgroup_size = None

        self.events = []
        self.kpsize = size
        self.buffers = {}
        self.programs = {}
        self.memory = None
        self.octave_max = None
        self.red_size = None
        if context:
            self.ctx = context
            device_name = self.ctx.devices[0].name.strip()
            platform_name = self.ctx.devices[0].platform.name.strip()
            platform = ocl.get_platform(platform_name)
            device = platform.get_device(device_name)
            self.device = platform.id, device.id
        else:
            if device is None:
                self.device = ocl.select_device(type=devicetype, memory=self.memory, best=True)
            else:
                self.device = device
            self.ctx = pyopencl.Context(devices=[pyopencl.get_platforms()[self.device[0]].get_devices()[self.device[1]]])
        if profile:
            self.queue = pyopencl.CommandQueue(self.ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
        else:
            self.queue = pyopencl.CommandQueue(self.ctx)
#        self._calc_workgroups()
        self._compile_kernels()
        self._allocate_buffers()
        self.debug = []
        self._sem = threading.Semaphore()

        self.devicetype = ocl.platforms[self.device[0]].devices[self.device[1]].type
        if (self.devicetype == "CPU"):
            self.USE_CPU = True
            self.matching_kernel = "matching_cpu"
            if sys.platform == "darwin":
                logger.warning("MacOSX computer working on CPU: limiting workgroup size to 1 !")
                self.max_workgroup_size = 1
                self.kernels = {}
                for k, v in self.__class__.kernels.items():
                    if isinstance(v, int):
                        self.kernels[k] = 1
                    else:
                        self.kernels[k] = tuple([1 for i in v])
        else:
            self.USE_CPU = False
            self.matching_kernel = "matching_gpu"
        self.roi = None
        if roi:
            self.set_roi(roi)

    def __del__(self):
        """
        Destructor: release all buffers
        """
        self._free_kernels()
        self._free_buffers()
        self.queue = None
        self.ctx = None
        gc.collect()

    def _allocate_buffers(self):
        self.buffers[ "Kp_1" ] = pyopencl.array.empty(self.queue, (self.kpsize,), dtype=self.dtype_kp)
        self.buffers[ "Kp_2" ] = pyopencl.array.empty(self.queue, (self.kpsize,), dtype=self.dtype_kp)
#        self.buffers[ "tmp" ] = pyopencl.array.empty(self.queue, (self.kpsize,), dtype=self.dtype_kp)
        self.buffers[ "match" ] = pyopencl.array.empty(self.queue, (self.kpsize, 2), dtype=numpy.int32)
        self.buffers["cnt" ] = pyopencl.array.empty(self.queue, 1, dtype=numpy.int32)

    def _free_buffers(self):
        """
        free all memory allocated on the device
        """
        for buffer_name in self.buffers:
            if self.buffers[buffer_name] is not None:
                try:
                    del self.buffers[buffer_name]
                    self.buffers[buffer_name] = None
                except pyopencl.LogicError:
                    logger.error("Error while freeing buffer %s" % buffer_name)

    def _compile_kernels(self):
        """
        Call the OpenCL compiler
        """
        kernel_directory = os.path.dirname(os.path.abspath(__file__))
        if not os.path.exists(os.path.join(kernel_directory, "matching_gpu" + ".cl")):
            while (".zip" in kernel_directory)  and (len(kernel_directory) > 4):
                kernel_directory = os.path.dirname(kernel_directory)
            kernel_directory = os.path.join(kernel_directory, "sift_kernels")

        for kernel in self.kernels:
            kernel_file = os.path.join(kernel_directory, kernel + ".cl")
            kernel_src = open(kernel_file).read()
            wg_size = self.kernels[kernel]
            try:
                program = pyopencl.Program(self.ctx, kernel_src).build('-D WORKGROUP_SIZE=%s' % wg_size)
            except pyopencl.MemoryError as error:
                raise MemoryError(error)
            except pyopencl.RuntimeError as error:
                if kernel == "keypoints":
                    logger.warning("Failed compiling kernel '%s' with workgroup size %s: %s: use low_end alternative", kernel, wg_size, error)
                    self.LOW_END = True
                else:
                    logger.error("Failed compiling kernel '%s' with workgroup size %s: %s", kernel, wg_size, error)
                    raise error
            self.programs[kernel] = program

    def _free_kernels(self):
        """
        free all kernels
        """
        self.programs = {}

    def match(self, nkp1, nkp2, raw_results=False):
        """
        calculate the matching of 2 keypoint list

        @param nkp1, nkp2: numpy 1D recarray of keypoints or equivalent GPU buffer
        @param raw_results: if true return the 2D array of indexes of matching keypoints (not the actual keypoints)

        TODO: implement the ROI ...

        """
        assert len(nkp1.shape) == 1  # Nota: nkp1.ndim is not valid for gpu_arrays
        assert len(nkp2.shape) == 1
        assert type(nkp1) in [numpy.core.records.recarray, pyopencl.array.Array]
        assert type(nkp2) in [numpy.core.records.recarray, pyopencl.array.Array]
        result = None
        with self._sem:
            if type(nkp1) == numpy.core.records.recarray:
                if nkp1.size > self.buffers[ "Kp_1" ].size:
                    logger.warning("increasing size of keypoint vector 1 to %i" % nkp1.size)
                    self.buffers[ "Kp_1" ] = pyopencl.array.empty(self.queue, (nkp1.size,), dtype=self.dtype_kp)
                kpt1_gpu = self.buffers[ "Kp_1" ]
                self._reset_buffer1()
                evt1 = pyopencl.enqueue_copy(self.queue, kpt1_gpu.data, nkp1)
                if self.profile:
                    self.events.append(("copy H->D KP_1", evt1))
            else:
                kpt1_gpu = nkp1
            if type(nkp2) == numpy.core.records.recarray:
                if nkp2.size > self.buffers[ "Kp_2" ].size:
                    logger.warning("increasing size of keypoint vector 2 to %i" % nkp2.size)
                    self.buffers[ "Kp_2" ] = pyopencl.array.empty(self.queue, (nkp2.size,), dtype=self.dtype_kp)
                kpt2_gpu = self.buffers[ "Kp_2" ]
                self._reset_buffer2()
                evt2 = pyopencl.enqueue_copy(self.queue, kpt2_gpu.data, nkp2)
                if self.profile:
                    self.events.append(("copy H->D KP_1", evt1))
            else:
                kpt2_gpu = nkp2
            if min(kpt1_gpu.size, kpt2_gpu.size) > self.buffers[ "match" ].shape[0]:
                self.kpsize = min(kpt1_gpu.size, kpt2_gpu.size)
                self.buffers[ "match" ] = pyopencl.array.empty(self.queue, (self.kpsize, 2), dtype=numpy.int32)
            self._reset_output()
            evt = self.programs[self.matching_kernel].matching(self.queue, calc_size((nkp1.size,), (self.kernels[self.matching_kernel],)), (self.kernels[self.matching_kernel],),
                                          kpt1_gpu.data,
                                          kpt2_gpu.data,
                                          self.buffers[ "match" ].data,
                                          self.buffers[ "cnt" ].data,
                                          numpy.int32(self.kpsize),
                                          numpy.float32(par.MatchRatio * par.MatchRatio),
                                          numpy.int32(nkp1.size),
                                          numpy.int32(nkp2.size))
            if self.profile:
                self.events.append(("matching", evt))
            size = self.buffers["cnt"].get()[0]
            match = numpy.empty(shape=(size, 2), dtype=numpy.int32)
            cpyD2H = pyopencl.enqueue_copy(self.queue, match, self.buffers[ "match" ].data)
            if self.profile:
                self.events.append(("copy D->H match", cpyD2H))
            if raw_results:
                result = match
            else:
                result = numpy.recarray(shape=(size, 2), dtype=self.dtype_kp)

                result[:, 0] = nkp1[match[:size, 0]]
                result[:, 1] = nkp2[match[:size, 1]]
        return result


    def _reset_buffer(self):
        """Reseet all buffers"""
        self._reset_buffer1()
        self._reset_buffer2()
        self._reset_output()

    def _reset_buffer1(self):
        ev1 = self.programs["memset"].memset_kp(self.queue, calc_size((self.buffers[ "Kp_1" ].size,), (self.kernels["memset"],)), (self.kernels["memset"],),
                                          self.buffers[ "Kp_1" ].data, numpy.float32(-1.0), numpy.uint8(0), numpy.int32(self.buffers[ "Kp_1" ].size))
        if self.profile:
            self.events.append(("memset Kp1", ev1))
    def _reset_buffer2(self):
        ev2 = self.programs["memset"].memset_kp(self.queue, calc_size((self.buffers[ "Kp_2" ].size,), (self.kernels["memset"],)), (self.kernels["memset"],),
                                          self.buffers[ "Kp_2" ].data, numpy.float32(-1.0), numpy.uint8(0), numpy.int32(self.buffers[ "Kp_2" ].size))
        if self.profile:
            self.events.append(("memset Kp2", ev2))
    def _reset_output(self):
        ev3 = self.programs["memset"].memset_int(self.queue, calc_size((self.buffers[ "match" ].size,), (self.kernels["memset"],)), (self.kernels["memset"],),
                                                 self.buffers[ "match" ].data, numpy.int32(-1), numpy.int32(self.buffers[ "match" ].size))
        ev4 = self.programs["memset"].memset_int(self.queue, (1,), (1,),
                                                 self.buffers[ "cnt" ].data, numpy.int32(0), numpy.int32(1))
        if self.profile:
            self.events += [("memset match", ev3),
                            ("memset cnt", ev4), ]
    def reset_timer(self):
        """
        Resets the profiling timers
        """
        with self._sem:
            self.events = []

    def set_roi(self, roi):
        """
        Defines the region of interest

        @param roi: region of interest as 2D numpy array with non zero where
        valid pixels are.
        """
        with self._sem:
            self.roi = numpy.ascontiguousarray(roi, numpy.int8)
            self.buffers["ROI"] = pyopencl.array.to_device(self.queue, self.roi)
    def unset_roi(self):
        """
        Unset the region of interest
        """
        with self._sem:
            self.roi = None
            self.buffers["ROI"] = None





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for image kernels
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc, scipy.ndimage, pylab
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
from test_image_functions import * #for Python implementation of tested functions
from test_image_setup import *
import sift_pyocl as sift
from sift_pyocl.utils import calc_size
logger = getLogger(__file__)
if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

SHOW_FIGURES = False
PRINT_KEYPOINTS = True
#USE_CPU = False
USE_CPP_SIFT = False #use reference cplusplus implementation for descriptors comparison... not valid for (octsize,scale)!=(1,1)



#print "working on %s" % ctx.devices[0].name

'''
For Python implementation of tested functions, see "test_image_functions.py"
'''


class ParameterisedTestCase(unittest.TestCase):
    """ TestCase classes that want to be parameterised should
        inherit from this class.
        From Eli Bendersky's website
        http://eli.thegreenplace.net/2011/08/02/python-unit-testing-parametrized-test-cases/
    """
    def __init__(self, methodName='runTest', param=None):
        super(ParameterisedTestCase, self).__init__(methodName)
        self.param = param

    @staticmethod
    def parameterise(testcase_klass, param=None):
        """ Create a suite containing all tests taken from the given
            subclass, passing them the parameter 'param'.
        """
        testloader = unittest.TestLoader()
        testnames = testloader.getTestCaseNames(testcase_klass)
        suite = unittest.TestSuite()
        for name in testnames:
            suite.addTest(testcase_klass(name, param=param))
        return suite


class test_keypoints(ParameterisedTestCase):
    def setUp(self):
        self.abort = False
        for kernel_file in self.param:
            if "cpu" in kernel_file:
                self.USE_CPU = True
            else:
                self.USE_CPU = False
            if kernel_file.startswith("orient"):
                self.wg_orient = self.param[kernel_file]
                kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), kernel_file + ".cl")
                kernel_src = open(kernel_path).read()
                try:
                    self.program_orient = pyopencl.Program(ctx, kernel_src).build()
                except:
                    logger.warning("Failed to compile kernel '%s': aborting" % kernel_file)
                    self.abort = True
                    return
            elif kernel_file.startswith("keypoint"):
                self.wg_keypoint = self.param[kernel_file]
                kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), kernel_file + ".cl")
                kernel_src = open(kernel_path).read()
                try:
                    self.program_keypoint = pyopencl.Program(ctx, kernel_src).build()
                except:
                    logger.warning("Failed to compile kernel '%s': aborting" % kernel_file)
                    self.abort = True
                    return

    def tearDown(self):
        self.mat = None
        self.program = None




    def test_orientation(self):
        '''
        #tests keypoints orientation assignment kernel
        '''
        if self.abort:
            return
        #orientation_setup :
        keypoints, nb_keypoints, updated_nb_keypoints, grad, ori, octsize = orientation_setup()
        keypoints, compact_cnt = my_compact(numpy.copy(keypoints),nb_keypoints)
        updated_nb_keypoints = compact_cnt
        
#        if (USE_CPU):
#            print("Using CPU-optimized kernels")
#            wg = 1,
#            shape = keypoints.shape[0]*wg[0],
#        else:
#            wg = 128, #FIXME : have to choose it for histograms #wg = max(self.wg),
        wg = self.wg_orient
        shape = keypoints.shape[0] * wg[0],  #shape = calc_size(keypoints.shape, self.wg)
        
        gpu_keypoints = pyopencl.array.to_device(queue, keypoints)
        actual_nb_keypoints = numpy.int32(updated_nb_keypoints)
        print("Number of keypoints before orientation assignment : %s" % actual_nb_keypoints)

        gpu_grad = pyopencl.array.to_device(queue, grad)
        gpu_ori = pyopencl.array.to_device(queue, ori)
        orisigma = numpy.float32(1.5) #SIFT
        grad_height, grad_width = numpy.int32(grad.shape)
        keypoints_start = numpy.int32(0)
        keypoints_end = numpy.int32(actual_nb_keypoints)
        counter = pyopencl.array.to_device(queue, keypoints_end) #actual_nb_keypoints)

        t0 = time.time()
        k1 = self.program_orient.orientation_assignment(queue, shape, wg,
        	gpu_keypoints.data, gpu_grad.data, gpu_ori.data, counter.data,
        	octsize, orisigma, nb_keypoints, keypoints_start, keypoints_end, grad_width, grad_height)
        res = gpu_keypoints.get()
        cnt = counter.get()
        t1 = time.time()
        
        if (USE_CPP_SIFT):
            import feature
            sc = feature.SiftAlignment()
            ref2 = sc.sift(scipy.misc.lena()) #ref2.x, ref2.y, ref2.scale, ref2.angle, ref2.desc --- ref2[numpy.argsort(ref2.y)]).desc
            ref = ref2.angle
            kp_ref = numpy.empty((ref2.size, 4), dtype=numpy.float32)
            kp_ref[:, 0] = ref2.x
            kp_ref[:, 1] = ref2.y
            kp_ref[:, 2] = ref2.scale
            kp_ref[:, 3] = ref2.angle
            
        else:
            ref, updated_nb_keypoints = my_orientation(keypoints, nb_keypoints, keypoints_start, keypoints_end, grad, ori, octsize, orisigma)
       
        t2 = time.time()
        
        if (PRINT_KEYPOINTS):
#            print("Keypoints after orientation assignment :")
#            print res[numpy.argsort(res[0:cnt,1])][0:cnt+10,3] #res[0:compact_cnt]
            print " "
#            print kp_ref[0:cnt+10]
#            print "Showing error (NOTE: significant error at position (i) should have its opposite at (i+1))"
#            print res[numpy.argsort(res[0:compact_cnt,1])][0:compact_cnt,3] - ref[0:compact_cnt]

#        print("Total keypoints for kernel : %s -- For Python : %s \t [octsize = %s]" % (cnt, updated_nb_keypoints, octsize))
#        print("Opencl found %s keypoints (%s added)" %(cnt,cnt-compact_cnt))
        
        #sort to compare added keypoints
        upbound = min(cnt,updated_nb_keypoints)
        d1, d2, d3, d4 = keypoints_compare(ref[0:upbound], res[0:upbound]) 
        self.assert_(d1 < 1e-4, "delta_cols=%s" % (d1))
        self.assert_(d2 < 1e-4, "delta_rows=%s" % (d2))
        self.assert_(d3 < 1e-4, "delta_sigma=%s" % (d3))
        self.assert_(d4 < 1e-1, "delta_angle=%s" % (d4)) #orientation has a poor precision
        logger.info("delta_cols=%s" % d1)
        logger.info("delta_rows=%s" % d2)
        logger.info("delta_sigma=%s" % d3)
        logger.info("delta_angle=%s" % d4)
        
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Orientation assignment took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
        

    def test_descriptor(self):
        '''
        #tests keypoints descriptors creation kernel
        '''
        if self.abort:
            return

        #descriptor_setup :
        keypoints_o, nb_keypoints, actual_nb_keypoints, grad, ori, octsize = descriptor_setup()
        #keypoints should be a compacted vector of keypoints
        keypoints_o, compact_cnt = my_compact(numpy.copy(keypoints_o),nb_keypoints)
        actual_nb_keypoints = compact_cnt
        keypoints_start, keypoints_end = 0, actual_nb_keypoints
        keypoints = keypoints_o[keypoints_start:keypoints_end+52] #to check if we actually stop at keypoints_end
        print("Working on keypoints : [%s,%s] (octave = %s)" % (keypoints_start, keypoints_end-1,int(numpy.log2(octsize)+1)))
        if not(USE_CPP_SIFT) and (100 < keypoints_end-keypoints_start): print "NOTE: Python implementation of descriptors is slow. Do not handle more than 100 keypoints, or grab a coffee..."
        
#        if (self.USE_CPU):
#            print "Using CPU-optimized kernels"
#            wg = 1,
#            shape = keypoints.shape[0]*wg[0],
#        else:
#            wg = (8, 8, 8)
#            shape = int(keypoints.shape[0]*wg[0]), 8, 8
#            wg = (4, 4, 8)
#            shape = int(keypoints.shape[0]*wg[0]), 4, 8
        wg = self.wg_keypoint
        if len(wg) == 1:
            shape = keypoints.shape[0] * wg[0],
        else:
            shape = keypoints.shape[0] * wg[0], wg[1], wg[2]
        gpu_keypoints = pyopencl.array.to_device(queue, keypoints)
        #NOTE: for the following line, use pyopencl.array.empty instead of pyopencl.array.zeros if the keypoints are compacted
        gpu_descriptors = pyopencl.array.zeros(queue, (keypoints_end - keypoints_start, 128), dtype=numpy.uint8, order="C")
        gpu_grad = pyopencl.array.to_device(queue, grad)
        gpu_ori = pyopencl.array.to_device(queue, ori)

        keypoints_start, keypoints_end = numpy.int32(keypoints_start), numpy.int32(keypoints_end)
        grad_height, grad_width = numpy.int32(grad.shape)
        counter = pyopencl.array.to_device(queue, keypoints_end)
        
        t0 = time.time()
        k1 = self.program_keypoint.descriptor(queue, shape, wg,
            gpu_keypoints.data, gpu_descriptors.data, gpu_grad.data, gpu_ori.data, numpy.int32(octsize),
            keypoints_start, counter.data, grad_width, grad_height)
        try:
            res = gpu_descriptors.get()
        except (pyopencl.LogicError, RuntimeError) as error:
            logger.warning("Segmentation fault like error (%s) on Descriptor for %s" % (error, self.param))
            return
        t1 = time.time()

        if (USE_CPP_SIFT):
            import feature
            sc = feature.SiftAlignment()
            ref2 = sc.sift(scipy.misc.lena()) #ref2.x, ref2.y, ref2.scale, ref2.angle, ref2.desc --- ref2[numpy.argsort(ref2.y)]).desc
            ref = ref2.desc
            ref_sort = ref
        else:
            ref = my_descriptor(keypoints_o, grad, ori, octsize, keypoints_start, keypoints_end)
            ref_sort = ref[numpy.argsort(keypoints[keypoints_start:keypoints_end,1])]

        t2 = time.time()
        
        if (PRINT_KEYPOINTS):
            res_sort = res[numpy.argsort(keypoints[keypoints_start:keypoints_end,1])]
            print res_sort[5:10]#keypoints_end-keypoints_start,0:15]
#            print res_sort[9]
            print ""
            print ref_sort[5:10]
#            numpy.savetxt("grrr_ocl_4_3.txt",res_sort,fmt='%d')
#            numpy.savetxt("grrr_cpp_4_3.txt",ref_sort,fmt='%d')
#            print ref[50:80,0:15]#[0:keypoints_end-keypoints_start,0:15]
            if (USE_CPP_SIFT and octsize == 1) or not(USE_CPP_SIFT): #this comparison is only relevant for the first keypoints
                print "Comparing descriptors (OpenCL and cpp) :"
                match, nulldesc = descriptors_compare(ref[keypoints_start:keypoints_end],res)
                print ("%s/%s match found" %(match,(keypoints_end-keypoints_start)-nulldesc))
#            print ref[1,:]
#            print res[1,:].sum(), ref[1,:].sum()

             #append to existing text file
#            f_handle = file('desc_by_test_keypoints.txt', 'a')
#            numpy.savetxt(f_handle,res_sort,fmt='%d')
#            f_handle.close()


        '''
            For now, the descriptor kernel is not precise enough to get exactly the same descriptors values 
        (we have several difference of 1, but it is OK for the SIFT matching).
            Use descriptors_compare(ref,res) to count how many descriptors are exactly the same.
        
        #sort to compare added keypoints
        delta = abs(res_sort-ref_sort).max()
        self.assert_(delta <= 1, "delta=%s" % (delta))
        logger.info("delta=%s" % delta)
        '''

        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Descriptors computation took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            
            
            
            

def test_suite_keypoints():
    testSuite = unittest.TestSuite()
    TESTCASES = [{"orientation_gpu":(128,), "keypoints_gpu2":(8, 8, 8)},
                 {"orientation_cpu":(1,), "keypoints_cpu":(1,)},
                  {"orientation_gpu":(128,), "keypoints_gpu1":(4, 4, 8)},

                  ]
    for param in TESTCASES:
        testSuite.addTest(ParameterisedTestCase.parameterise(
                test_keypoints, param))

#    testSuite.addTest(test_keypoints("test_orientation"))
#    testSuite.addTest(test_keypoints("test_descriptor"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_keypoints()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for all sift_pyocl
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time,os
import sys
import unittest
from utilstest import UtilsTest, getLogger
logger = getLogger(__file__)

from test_gaussian import test_suite_gaussian
from test_preproc import test_suite_preproc
from test_reductions import test_suite_reductions
from test_convol import test_suite_convol
from test_algebra import test_suite_algebra
from test_image import test_suite_image
from test_keypoints import test_suite_keypoints
from test_matching import test_suite_matching

def test_suite_all():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_suite_algebra())
    testSuite.addTest(test_suite_gaussian())
    testSuite.addTest(test_suite_preproc())
    testSuite.addTest(test_suite_reductions())
    testSuite.addTest(test_suite_convol())
    testSuite.addTest(test_suite_image())
    testSuite.addTest(test_suite_keypoints())
    testSuite.addTest(test_suite_matching())
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_all()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import numpy
from pycuda import driver, compiler, gpuarray, tools

# -- initialize the device
import pycuda.autoinit
import scipy.misc

kernel_global = """
__global__ void conv_global(float *inp, float *out, int width, int height)
{
    const uint x = blockIdx.x * blockDim.x + threadIdx.x;
    const uint y = blockIdx.y * blockDim.y + threadIdx.y;

    uint xplus, yplus;
    if (x<width-1) xplus=x+1;
    else           xplus=x;
    if (y<height-1)yplus=y+1;
    else           yplus=y;


    out[y*width+x] = 0.25f*(inp[y*width+x]+
                            inp[yplus*width+x]+
                            inp[y*width+xplus]+
                            inp[yplus*width+xplus]);
}


__global__ void conv_shared(float *inp, float *out, int width, int height)
{
    const uint shmWidth = blockDim.x + 1;
    const uint x = blockIdx.x * blockDim.x + threadIdx.x;
    const uint y = blockIdx.y * blockDim.y + threadIdx.y;
    __shared__ float shm[%s];

    shm[ shmWidth*threadIdx.y+threadIdx.x ] = inp[ y*width+x ];
    if (threadIdx.x==blockDim.x-1){
        if (x == width-1)
            shm[ shmWidth*threadIdx.y+threadIdx.x+1 ] = shm[ shmWidth*threadIdx.y+threadIdx.x ];
        else
            shm[ shmWidth*threadIdx.y+threadIdx.x+1 ] = inp[ y*width+x+1 ];
        }
    if (threadIdx.y==blockDim.y-1){
        if (y == height-1)
            shm[ shmWidth*(threadIdx.y+1)+threadIdx.x ] = shm[ shmWidth*threadIdx.y+threadIdx.x ];
        else
            shm[ shmWidth*(threadIdx.y+1)+threadIdx.x ] = inp[ (y+1)*width+x ];
        }
    if ((threadIdx.x==blockDim.x-1)&&(threadIdx.y==blockDim.y-1)){
        if ((x == width-1) && (y == height-1))
            shm[ shmWidth*(threadIdx.y+1)+threadIdx.x+1 ] = shm[ shmWidth*threadIdx.y+threadIdx.x ];
        else
            shm[ shmWidth*(threadIdx.y+1)+threadIdx.x+1 ] = inp[ (y+1)*width+x+1 ];
        }
    __syncthreads();

    out[y*width+x] = 0.25f*(shm[threadIdx.y*shmWidth+threadIdx.x]+
                            shm[(threadIdx.y+1)*shmWidth+threadIdx.x]+
                            shm[threadIdx.y*shmWidth+threadIdx.x+1]+
                            shm[(threadIdx.y+1)*shmWidth+threadIdx.x+1]);
}

"""



BS = 16
# create two random square matrices
a_cpu = scipy.misc.lena().astype("float32")
height, width = a_cpu.shape

aplus = numpy.empty((height+1,width+1),numpy.float32)
aplus[1:,1:]=a_cpu
aplus[:-1,:-1]=a_cpu
ref = 0.25 * (aplus[:-1, :-1] + aplus[1:, :-1] + aplus[:-1, 1:] + aplus[1:, 1:])

# transfer host (CPU) memory to device (GPU) memory
a_gpu = gpuarray.to_gpu(a_cpu)
b_gpu = gpuarray.empty_like(a_gpu)

kernel_code = kernel_global % ((BS + 1) * (BS + 1))

# compile the kernel code
mod = compiler.SourceModule(kernel_code)

# get the kernel function from the compiled module
conv_global = mod.get_function("conv_global")
conv_shared = mod.get_function("conv_shared")
# call the kernel on the card
t0 = time.time()
conv_global(
    # inputs
    a_gpu, b_gpu,
    numpy.int32(width),
    numpy.int32(height),
    # grid of multiple blocks
    grid=(width / BS, height / BS),
    # block of multiple threads
    block=(BS, BS, 1),
    )
print "Global", 1e6 * (time.time() - t0)
b_cpu = b_gpu.get()
numpy.allclose(ref, b_cpu)

t0 = time.time()
conv_shared(
    # inputs
    a_gpu, b_gpu,
    numpy.int32(width),
    numpy.int32(height),
    # grid of multiple blocks
    grid=(width / BS, height / BS),
    # block of multiple threads
    block=(BS, BS, 1),
    )
print "Shared", 1e6 * (time.time() - t0)
b_cpu = b_gpu.get()
numpy.allclose(ref, b_cpu)
if 1:
    import pylab
    pylab.ion()
    fig = pylab.figure()
    sp1 = fig.add_subplot(221)
    sp2 = fig.add_subplot(222)
    sp3 = fig.add_subplot(223)
    sp4 = fig.add_subplot(224)
    sp1.imshow(a_cpu, cmap="gray", interpolation="nearest")
    sp2.imshow(ref, cmap="gray", interpolation="nearest")
    sp3.imshow(b_cpu, cmap="gray", interpolation="nearest")
    sp4.imshow(ref - b_cpu, cmap="gray", interpolation="nearest")
    fig.show()
    raw_input("enter")




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Contains classes for image alignment on a reference images.
"""

from __future__ import division, print_function, with_statement

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-07-24"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import os, gc, sys
from threading import Semaphore
import numpy
from .param import par
from .opencl import ocl, pyopencl
from .utils import calc_size, kernel_size, sizeof, matching_correction
import logging
logger = logging.getLogger("sift.alignment")
if pyopencl:
    from pyopencl import mem_flags as MF
else:
    logger.warning("No PyOpenCL, no sift")

from . import MatchPlan, SiftPlan

try:
    import feature
except ImportError:
    feature = None

def arrow_start(kplist):
    x_ref = kplist.x
    y_ref = kplist.y
    angle_ref = kplist.angle
    scale_ref = kplist.scale
    x_ref2 = kplist.x + scale_ref * numpy.cos(angle_ref)
    y_ref2 = kplist.y + scale_ref * numpy.sin(angle_ref)
    return x_ref2, y_ref2
def transform_pts(matrix, offset, x, y):
    nx = -offset[1] + y * matrix[1, 0] + x * matrix[1, 1]
    ny = -offset[0] + x * matrix[0, 1] + y * matrix[0, 0]
    return nx, ny

class LinearAlign(object):
    """
    Align images on a reference image based on an Afine transformation (bi-linear + offset)
    """
    kernels = {"transform":128}

    def __init__(self, image, devicetype="CPU", profile=False, device=None, max_workgroup_size=None,
                 ROI=None, extra=0, context=None, init_sigma=None):
        """
        Constructor of the class

        @param image: reference image on which other image should be aligned
        @param devicetype: Kind of preferred devce
        @param profile:collect profiling information ?
        @param device: 2-tuple of integer. see clinfo
        @param max_workgroup_size: set to 1 for macOSX on CPU
        @param ROI: Region of interest: to be implemented
        @param extra: extra space around the image, can be an integer, or a 2 tuple in YX convention: TODO!
        @param init_sigma: bluring width, you should have good reasons to modify the 1.6 default value...
        """
        self.profile = bool(profile)
        self.events = []
        self.program = None
        self.ref = numpy.ascontiguousarray(image, numpy.float32)
        self.buffers = {}
        self.shape = image.shape
        if max_workgroup_size:
            self.max_workgroup_size = int(max_workgroup_size)
            self.kernels = {}
            for k, v in self.__class__.kernels.items():
                self.kernels[k] = min(v, self.max_workgroup_size)
        else:
            self.max_workgroup_size = None
        if len(self.shape) == 3:
            self.RGB = True
            self.shape = self.shape[:2]
        elif len(self.shape) == 2:
            self.RGB = False
        else:
            raise RuntimeError("Unable to process image of shape %s" % (tuple(self.shape,)))
        if "__len__" not in dir(extra):
            self.extra = (int(extra), int(extra))
        else:
            self.extra = extra[:2]
        self.outshape = tuple(i + 2 * j for i, j in zip(self.shape, self.extra))
        self.ROI = ROI
        if context:
            self.ctx = context
            device_name = self.ctx.devices[0].name.strip()
            platform_name = self.ctx.devices[0].platform.name.strip()
            platform = ocl.get_platform(platform_name)
            device = platform.get_device(device_name)
            self.device = platform.id, device.id
        else:
            if device is None:
                self.device = ocl.select_device(type=devicetype, best=True)
            else:
                self.device = device
            self.ctx = pyopencl.Context(devices=[pyopencl.get_platforms()[self.device[0]].get_devices()[self.device[1]]])
        self.devicetype = ocl.platforms[self.device[0]].devices[self.device[1]].type
        if (self.devicetype == "CPU"):
            self.USE_CPU = True
            if sys.platform == "darwin":
                logger.warning("MacOSX computer working on CPU: limiting workgroup size to 1 !")
                self.max_workgroup_size = 1
                self.kernels = {}
                for k, v in self.__class__.kernels.items():
                    if isinstance(v, int):
                        self.kernels[k] = 1
                    else:
                        self.kernels[k] = tuple([1 for i in v])
        if self.RGB:
            if self.max_workgroup_size == 1:
                self.wg = (1, 1, 1)
            else:
                self.wg = (4, 8, 4)
        else:
            if self.max_workgroup_size == 1:
                self.wg = (1, 1)
            else:
                self.wg = (8, 4)


        self.sift = SiftPlan(template=image, context=self.ctx, profile=self.profile,
                             max_workgroup_size=self.max_workgroup_size, init_sigma=init_sigma)
        self.ref_kp = self.sift.keypoints(image)
        if self.ROI is not None:
            kpx = numpy.round(self.ref_kp.x).astype(numpy.int32)
            kpy = numpy.round(self.ref_kp.y).astype(numpy.int32)
            masked = self.ROI[(kpy, kpx)].astype(bool)
            logger.warning("Reducing keypoint list from %i to %i because of the ROI" % (self.ref_kp.size, masked.sum()))
            self.ref_kp = self.ref_kp[masked]
        self.match = MatchPlan(context=self.ctx, profile=self.profile, max_workgroup_size=self.max_workgroup_size)
#        Allocate reference keypoints on the GPU within match context:
        self.buffers["ref_kp_gpu"] = pyopencl.array.to_device(self.match.queue, self.ref_kp)
        # TODO optimize match so that the keypoint2 can be optional
        self.fill_value = 0
#        print self.ctx.devices[0]
        if self.profile:
            self.queue = pyopencl.CommandQueue(self.ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
        else:
            self.queue = pyopencl.CommandQueue(self.ctx)
        self._compile_kernels()
        self._allocate_buffers()
        self.sem = Semaphore()
        self.relative_transfo = None

    def __del__(self):
        """
        Destructor: release all buffers
        """
        self._free_kernels()
        self._free_buffers()
        self.queue = None
        self.ctx = None
        gc.collect()

    def _allocate_buffers(self):
        """
        All buffers are allocated here
        """
        if self.RGB:
            self.buffers["input"] = pyopencl.array.empty(self.queue, shape=self.shape + (3,), dtype=numpy.uint8)
            self.buffers["output"] = pyopencl.array.empty(self.queue, shape=self.outshape + (3,), dtype=numpy.uint8)
        else:
            self.buffers["input"] = pyopencl.array.empty(self.queue, shape=self.shape, dtype=numpy.float32)
            self.buffers["output"] = pyopencl.array.empty(self.queue, shape=self.outshape, dtype=numpy.float32)
        self.buffers["matrix"] = pyopencl.array.empty(self.queue, shape=(2, 2), dtype=numpy.float32)
        self.buffers["offset"] = pyopencl.array.empty(self.queue, shape=(1, 2), dtype=numpy.float32)
    def _free_buffers(self):
        """
        free all memory allocated on the device
        """
        for buffer_name in self.buffers:
            if self.buffers[buffer_name] is not None:
                try:
                    del self.buffers[buffer_name]
                    self.buffers[buffer_name] = None
                except pyopencl.LogicError:
                    logger.error("Error while freeing buffer %s" % buffer_name)

    def _compile_kernels(self):
        """
        Call the OpenCL compiler
        """
        kernel_directory = os.path.dirname(os.path.abspath(__file__))
        kernel_file = self.kernels.keys()[0]
        if not os.path.exists(os.path.join(kernel_directory, kernel_file + ".cl")):
            while (".zip" in kernel_directory)  and (len(kernel_directory) > 4):
                kernel_directory = os.path.dirname(kernel_directory)
            kernel_directory = os.path.join(kernel_directory, "sift_kernels")
        kernel_file = os.path.join(kernel_directory, kernel_file + ".cl")
        kernel_src = open(kernel_file).read()
        try:
            program = pyopencl.Program(self.ctx, kernel_src).build()
        except pyopencl.MemoryError as error:
            raise MemoryError(error)
        self.program = program

    def _free_kernels(self):
        """
        free all kernels
        """
        self.program = None

    def align(self, img, shift_only=False, return_all=False, double_check=False, relative=False, orsa=False):
        """
        Align image on reference image

        @param img: numpy array containing the image to align to reference
        @param return_all: return in addition ot the image, keypoints, matching keypoints, and transformations as a dict
        @param reltive: update reference keypoints with those from current image to perform relative alignment
        @return: aligned image or all informations
        """
        logger.debug("ref_keypoints: %s" % self.ref_kp.size)
        if self.RGB:
            data = numpy.ascontiguousarray(img, numpy.uint8)
        else:
            data = numpy.ascontiguousarray(img, numpy.float32)
        with self.sem:
            cpy = pyopencl.enqueue_copy(self.queue, self.buffers["input"].data, data)
            if self.profile:self.events.append(("Copy H->D", cpy))
            cpy.wait()
            kp = self.sift.keypoints(self.buffers["input"])
#            print("ref %s img %s" % (self.buffers["ref_kp_gpu"].shape, kp.shape))
            logger.debug("mod image keypoints: %s" % kp.size)
            raw_matching = self.match.match(self.buffers["ref_kp_gpu"], kp, raw_results=True)
#            print(raw_matching.max(axis=0))

            matching = numpy.recarray(shape=raw_matching.shape, dtype=MatchPlan.dtype_kp)
            len_match = raw_matching.shape[0]
            if len_match == 0:
                logger.warning("No matching keypoints")
                return
            matching[:, 0] = self.ref_kp[raw_matching[:, 0]]
            matching[:, 1] = kp[raw_matching[:, 1]]

            if orsa:
                if feature:
                    matching = feature.sift_orsa(matching, self.shape, 1)
                else:
                    logger.warning("feature is not available. No ORSA filtering")

            if (len_match < 3 * 6) or (shift_only):  # 3 points per DOF
                if shift_only:
                    logger.debug("Shift Only mode: Common keypoints: %s" % len_match)
                else:
                    logger.warning("Shift Only mode: Common keypoints: %s" % len_match)
                dx = matching[:, 1].x - matching[:, 0].x
                dy = matching[:, 1].y - matching[:, 0].y
                matrix = numpy.identity(2, dtype=numpy.float32)
                offset = numpy.array([+numpy.median(dy), +numpy.median(dx)], numpy.float32)
            else:
                logger.debug("Common keypoints: %s" % len_match)

                transform_matrix = matching_correction(matching)
                offset = numpy.array([transform_matrix[5], transform_matrix[2]], dtype=numpy.float32)
                matrix = numpy.empty((2, 2), dtype=numpy.float32)
                matrix[0, 0], matrix[0, 1] = transform_matrix[4], transform_matrix[3]
                matrix[1, 0], matrix[1, 1] = transform_matrix[1], transform_matrix[0]
            if double_check and (len_match >= 3 * 6):  # and abs(matrix - numpy.identity(2)).max() > 0.1:
                logger.warning("Validating keypoints, %s,%s" % (matrix, offset))
                dx = matching[:, 1].x - matching[:, 0].x
                dy = matching[:, 1].y - matching[:, 0].y
                dangle = matching[:, 1].angle - matching[:, 0].angle
                dscale = numpy.log(matching[:, 1].scale / matching[:, 0].scale)
                distance = numpy.sqrt(dx * dx + dy * dy)
                outlayer = numpy.zeros(distance.shape, numpy.int8)
                outlayer += abs((distance - distance.mean()) / distance.std()) > 4
                outlayer += abs((dangle - dangle.mean()) / dangle.std()) > 4
                outlayer += abs((dscale - dscale.mean()) / dscale.std()) > 4
                print(outlayer)
                outlayersum = outlayer.sum()
                if outlayersum > 0 and not numpy.isinf(outlayersum):
                    matching2 = matching[outlayer == 0]
                    transform_matrix = matching_correction(matching2)
                    offset = numpy.array([transform_matrix[5], transform_matrix[2]], dtype=numpy.float32)
                    matrix = numpy.empty((2, 2), dtype=numpy.float32)
                    matrix[0, 0], matrix[0, 1] = transform_matrix[4], transform_matrix[3]
                    matrix[1, 0], matrix[1, 1] = transform_matrix[1], transform_matrix[0]
            if relative:  # update stable part to perform a relative alignment
                self.ref_kp = kp
                if self.ROI is not None:
                    kpx = numpy.round(self.ref_kp.x).astype(numpy.int32)
                    kpy = numpy.round(self.ref_kp.y).astype(numpy.int32)
                    masked = self.ROI[(kpy, kpx)].astype(bool)
                    logger.warning("Reducing keypoint list from %i to %i because of the ROI" % (self.ref_kp.size, masked.sum()))
                    self.ref_kp = self.ref_kp[masked]
                self.buffers["ref_kp_gpu"] = pyopencl.array.to_device(self.match.queue, self.ref_kp)
                transfo = numpy.zeros((3, 3), dtype=numpy.float64)
                transfo[:2, :2] = matrix
                transfo[0, 2] = offset[0]
                transfo[1, 2] = offset[1]
                transfo[2, 2] = 1
                if self.relative_transfo is None:
                    self.relative_transfo = transfo
                else:
                    self.relative_transfo = numpy.dot(transfo, self.relative_transfo)
                matrix = numpy.ascontiguousarray(self.relative_transfo[:2, :2], dtype=numpy.float32)
                offset = numpy.ascontiguousarray(self.relative_transfo[:2, 2], dtype=numpy.float32)
#                print(self.relative_transfo)
            cpy1 = pyopencl.enqueue_copy(self.queue, self.buffers["matrix"].data, matrix)
            cpy2 = pyopencl.enqueue_copy(self.queue, self.buffers["offset"].data, offset)
            if self.profile:
                self.events += [("Copy matrix", cpy1), ("Copy offset", cpy2)]

            if self.RGB:
                shape = (4, self.shape[1], self.shape[0])
                transform = self.program.transform_RGB
            else:
                shape = self.shape[1], self.shape[0]
                transform = self.program.transform
            ev = transform(self.queue, calc_size(shape, self.wg), self.wg,
                                   self.buffers["input"].data,
                                   self.buffers["output"].data,
                                   self.buffers["matrix"].data,
                                   self.buffers["offset"].data,
                                   numpy.int32(self.shape[1]),
                                   numpy.int32(self.shape[0]),
                                   numpy.int32(self.outshape[1]),
                                   numpy.int32(self.outshape[0]),
                                   self.sift.buffers["min"].get()[0],
                                   numpy.int32(1))
            if self.profile:
                self.events += [("transform", ev)]
            result = self.buffers["output"].get()

#        print (self.buffers["offset"])
        if return_all:
#            corr = numpy.dot(matrix, numpy.vstack((matching[:, 1].y, matching[:, 1].x))).T - \
#                   offset.T - numpy.vstack((matching[:, 0].y, matching[:, 0].x)).T
            corr = numpy.dot(matrix, numpy.vstack((matching[:, 0].y, matching[:, 0].x))).T + offset.T - numpy.vstack((matching[:, 1].y, matching[:, 1].x)).T
            rms = numpy.sqrt((corr * corr).sum(axis=-1).mean())

            # Todo: calculate the RMS of deplacement and return it:
            return {"result":result, "keypoint":kp, "matching":matching, "offset":offset, "matrix": matrix, "rms":rms}
        return result

    def log_profile(self):
        """
        If we are in debugging mode, prints out all timing for every single OpenCL call
        """
        t = 0.0
        orient = 0.0
        descr = 0.0
        if self.profile:
            for e in self.events:
                if "__len__" in dir(e) and len(e) >= 2:
                    et = 1e-6 * (e[1].profile.end - e[1].profile.start)
                    print("%50s:\t%.3fms" % (e[0], et))
                    t += et






#!/usr/bin/python
# -*- coding: utf-8 -*
"""
Small demonstration for sift using pyOpenCL
"""
from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-09-05"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import sys, os, pyopencl, time, urllib2
from math import sin, cos
import logging
logger = logging.getLogger("sift")
import sift_pyocl as sift
import numpy
import scipy.misc
import pylab
try:
    import feature
except:
    logger.error("Feature is not available to compare results with C++ implementation")
    feature = None
    res = numpy.empty(0)

def cmp_kp(a, b):
    if a.scale < b.scale:
        return True
    elif a.scale > b.scale:
        return False
    else:
        if a.angle > b.angle:
            return True
        else:
            return False


class DemoSift(object):
    def __init__(self, filename=None, devicetype=None, device=None, profile=False):
        if filename and os.path.exists(filename):
            self.filename = filename
        else:
            self.filename = "Esrf_grenoble.jpg"
            data = urllib2.urlopen("http://upload.wikimedia.org/wikipedia/commons/9/94/Esrf_grenoble.jpg").read()
            open(self.filename, "wb").write(data)
        self.image_rgb = scipy.misc.imread(self.filename)
        if self.image_rgb.ndim != 2:
            self.image_bw = 0.299 * self.image_rgb[:, :, 0] + 0.587 * self.image_rgb[:, :, 1] + 0.114 * self.image_rgb[:, :, 2]
        else: self.image_bw = self.image_rgb
        if feature:
            self._sift_cpp = feature.SiftAlignment()
        self._sift_ocl = sift.SiftPlan(template=self.image_rgb, device=device, devicetype=devicetype, profile=profile)
        self.kp_cpp = numpy.empty(0)
        self.kp_ocl = numpy.empty(0)
        self.fig = pylab.figure()
        self.sp1 = self.fig.add_subplot(1, 2, 1)
        self.sp2 = self.fig.add_subplot(1, 2, 2)
        #elf.sp3 = self.fig.add_subplot(1, 2, 3)
        #elf.sp4 = self.fig.add_subplot(1, 2, 4)

        self.im1 = self.sp1.imshow(self.image_rgb)
        self.sp1.set_title("OpenCL: %s keypoint" % self.kp_ocl.size)
        self.im2 = self.sp2.imshow(self.image_bw, cmap="gray")
        self.sp2.set_title("C++: %s keypoint" % self.kp_cpp.size)
        self.fig.show()
        self.timing_cpp = None
        self.timing_ocl = None

    def sift_cpp(self):
        print(os.linesep + "Running SIFT using C++ code")
        t0 = time.time()
        self.kp_cpp = self._sift_cpp.sift(self.image_bw)
        t1 = time.time()
        self.timing_cpp = t1 - t0
        if "size" not in dir(self.kp_cpp):
            return #we are using an old kind of Sift-C++
        self.sp2.set_title("C++: %s keypoint" % self.kp_cpp.size)
        self.fig.canvas.draw()
        self.kp_cpp.sort(order=["scale", "angle", "x", "y"])

    def sift_ocl(self):
        print(os.linesep + "Running SIFT using OpenCL code")
        t0 = time.time()
        self.kp_ocl = self._sift_ocl.keypoints(self.image_rgb)
        t1 = time.time()
        self.timing_ocl = t1 - t0
        self.sp1.set_title("OpenCL: %s keypoint" % self.kp_ocl.size)
        self.fig.canvas.draw()
        self.kp_ocl.sort(order=["scale", "angle", "x", "y"])

    def timings(self):
        if self.kp_ocl.size > 0 and self.kp_cpp > 0:
            print("Computing time using C++: %.3fms\t using OpenCL: %.3fms:\t Speed up: %.3f" % (1e3 * self.timing_cpp, 1e3 * self.timing_ocl, self.timing_cpp / self.timing_ocl))

    def show(self, max=1000):
        if max == None:
            max = sys.maxint
        if self.kp_cpp.size > max:
            print("keeping only the %i largest keypoints for display" % max)
        todo = min(self.kp_cpp.size, max)

        for i in range(self.kp_cpp.size - todo, self.kp_cpp.size):
            x = self.kp_cpp[i].x
            y = self.kp_cpp[i].y
            scale = self.kp_cpp[i].scale
            angle = self.kp_cpp[i].angle
            x0 = x + scale * cos(angle)
            y0 = y + scale * sin(angle)
            self.sp2.annotate("", xy=(x, y), xytext=(x0, y0), color="red",
                             arrowprops=dict(facecolor='red', edgecolor='red', width=1),)
            self.sp1.annotate("", xy=(x, y), xytext=(x0, y0), color="red",
                             arrowprops=dict(facecolor='red', edgecolor='red', width=1),)
        self.fig.canvas.draw()

        todo = min(self.kp_ocl.size, max)
        for i in range(self.kp_ocl.size - todo, self.kp_ocl.size):
            x = self.kp_ocl[i].x
            y = self.kp_ocl[i].y
            scale = self.kp_ocl[i].scale
            angle = self.kp_ocl[i].angle
            x0 = x + scale * cos(angle)
            y0 = y + scale * sin(angle)
            self.sp1.annotate("", xy=(x, y), xytext=(x0, y0), color="blue",
                             arrowprops=dict(facecolor='blue', edgecolor='blue', width=1),)
        self.fig.canvas.draw()

    def match(self):
        if self.kp_ocl.size > 0 and self.kp_cpp > 0:
            t0 = time.time()
            match = feature.sift_match(self.kp_ocl, self.kp_cpp)
            t1 = time.time()
            print("Number of Keypoints with OpenCL: %i, With C++: %i, Matching keypoints: %i, Processing time: %.3fs" % (self.kp_ocl.size, self.kp_cpp.size, match.shape[0], (t1 - t0)))
            self.fig.canvas.draw()
        else:
            print("No keypoints, cannot match")


if __name__ == "__main__":
    from optparse import OptionParser
    parser = OptionParser(version="1.0", description="Demonstration of sift using OpenCL version C++ implementation",
                          usage="usage: %prog [options] imagefiles*")

    parser.add_option("-d", "--device", dest="device",
                  help="device on which to run: coma separated like --device=0,1",
                  default=None)
    parser.add_option("-t", "--type", dest="type",
                       help="device type  on which to run like CPU (default) or GPU",
                       default="CPU")
    parser.add_option("-p", "--profile", dest="profile",
                       help="Print profiling information of OpenExecution",
                       default=False, action="store_true")
    parser.add_option("-f", "--force", dest="force",
                       help="rebuild the package",
                       default=False, action="store_true")
    parser.add_option("-r", "--remove", dest="remove",
                       help="remove build and rebuild the package",
                       default=False, action="store_true")
    (options, args) = parser.parse_args()
    if options.device:
        device = tuple(int(i) for i in options.device.split(","))
    else:
        device = None
    if args:
        for i in args:
            if os.path.exists(i):
                print("Processing file %s" % i)
                d = DemoSift(i, devicetype=options.type, device=device, profile=options.profile)
                d.sift_ocl()
                if feature:
                    d.sift_cpp()
                d.timings()
                d.show(1000)
                d.match()
                if options.profile:
                    d._sift_ocl.log_profile()
                raw_input()
    else:
        print("Processing file demo image")
        d = DemoSift(devicetype=options.type, device=device, profile=options.profile)
        d.sift_ocl()
        if feature:
            d.sift_cpp()
        d.timings()
        d.show(1000)
        d.match()
        if options.profile:
            d._sift_ocl.log_profile()
        raw_input()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#


from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-06-13"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

from math import ceil
import numpy

def calc_size(shape, blocksize):
    """
    Calculate the optimal size for a kernel according to the workgroup size
    """
    if "__len__" in dir(blocksize):
        return tuple((int(i) + int(j) - 1) & ~(int(j) - 1) for i, j in zip(shape, blocksize))
    else:
        return tuple((int(i) + int(blocksize) - 1) & ~(int(blocksize) - 1) for i in shape)


def kernel_size(sigma, odd=False, cutoff=4):
    """
    Calculate the optimal kernel size for a convolution with sigma

    @param sigma: width of the gaussian
    @param odd: enforce the kernel to be odd (more precise ?)
    """
    size = int(ceil(2 * cutoff * sigma + 1))
    if odd and size % 2 == 0:
        size += 1
    return size

def sizeof(shape, dtype="uint8"):
    """
    Calculate the number of bytes needed to allocate for a given structure

    @param shape: size or tuple of sizes
    @param dtype: data type
    """
    itemsize = numpy.dtype(dtype).itemsize
    cnt = 1
    if "__len__" in dir(shape):
        for dim in shape:
            cnt *= dim
    else:
        cnt = int(shape)
    return cnt * itemsize

def _gcd(a, b):
    """Calculate the greatest common divisor of a and b"""
    while b:
        a, b = b, a % b
    return a



def matching_correction(matching):
    '''
    Given the matching between two list of keypoints, 
    return the linear transformation to correct kp2 with respect to kp1
    '''
    N = matching.shape[0]
    #solving normals equations for least square fit
    X = numpy.zeros((2 * N, 6))
    X[::2, 2:] = 1, 0, 0, 0
    X[::2, 0] = matching.x[:, 0]
    X[::2, 1] = matching.y[:, 0]
    X[1::2, 0:3] = 0, 0, 0
    X[1::2, 3] = matching.x[:, 0]
    X[1::2, 4] = matching.y[:, 0]
    X[1::2, 5] = 1
    y = numpy.zeros((2 * N, 1))
    y[::2, 0] = matching.x[:, 1]
    y[1::2, 0] = matching.y[:, 1]
    A = numpy.dot(X.transpose(), X)
    sol = numpy.dot(numpy.linalg.inv(A), numpy.dot(X.transpose(), y))
#    sol = numpy.dot(numpy.linalg.pinv(X),y) #pseudo-inverse is slower
#    MSE = numpy.linalg.norm(y - numpy.dot(X,sol))**2/N #Mean Squared Error, if needed
    return sol




def bin2RGB(img):
    """
    Perform a 2x2 binning of the image
    """
    dtype = img.dtype
    if dtype == numpy.uint8:
        out_dtype = numpy.int32
    else:
        out_dtype = dtype
    shape = img.shape
    if len(shape) == 3:
        new_shape = shape[0] // 2, shape[1] // 2, shape[2]
        new_img = img
    else:
        new_shape = shape[0] // 2, shape[1] // 2, 1
        new_img = img.reshape((shape[0], shape[1], 1))
    out = numpy.zeros(new_shape, dtype=out_dtype)
    out += new_img[::2, ::2, :]
    out += new_img[1::2, ::2, :]
    out += new_img[1::2, 1::2, :]
    out += new_img[::2, 1::2, :]
    out /= 4
    if len(shape) != 3:
        out.shape = new_shape[0], new_shape[1]
    if dtype == numpy.uint8:
        return out.astype(dtype)
    else:
        return out













#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for alignment module
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-08-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc, scipy.ndimage
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
import sift_pyocl as sift
from sift_pyocl.alignment import LinearAlign
logger = getLogger(__file__)
if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    import pylab
else:
    PROFILE = False
PRINT_KEYPOINTS = False




class test_linalign(unittest.TestCase):
    def setUp(self):
        self.lena = scipy.misc.lena().astype(numpy.float32)
        self.shape = self.lena.shape
        self.extra = (10, 11)
#        self.img = scipy.ndimage.shift(self.lena, (7, 5))
#        self.img = scipy.ndimage.rotate(self.lena, -20, reshape=False, order=3)
#        self.img = scipy.ndimage.shift(scipy.ndimage.rotate(self.lena, 20, reshape=False, order=3), (7, 5))
        self.img = scipy.ndimage.affine_transform(self.lena, [[1.1, -0.1], [0.05, 0.9]], [7, 5])
        self.align = LinearAlign(self.lena, context=ctx)


    def test_align(self):
        """
        tests the combine (linear combination) kernel
        """
        out = self.align.align(self.img, 0, 1)
        for i in out:
            if i in ["offset","matrix"]:
                print i
                print out[i]
        self.align.log_profile()
        out = out["result"]

        if PROFILE and out is not None:
            fig = pylab.figure()
            sp0 = fig.add_subplot(221)
            im0 = sp0.imshow(self.lena)
            sp1 = fig.add_subplot(222)
            im1 = sp1.imshow(self.img)
            sp2 = fig.add_subplot(223)
            im2 = sp2.imshow(out)
            sp3 = fig.add_subplot(224)
            delta = (out - self.lena)[100:400, 100:400]
            im3 = sp3.imshow(delta)
            print({"min":delta.min(), "max:":delta.max(), "mean":delta.mean(), "std:":delta.std()})
            pylab.show()
            raw_input("enter")



def test_suite_align():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_linalign("test_align"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_align()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)




#!/usr/bin/python
#from utilstest import UtilsTest#, getLogger
#logger = getLogger(__file__)
import sift_pyocl as sift
import numpy
import scipy.misc

import sys, os, copy, time
#import numpy, scipy.misc
from PyQt4 import QtGui, QtCore


#TODO: remove this after integrating in sift
class Enum(dict):
    """
    Simple class half way between a dict and a class, behaving as an enum
    """
    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError

'''
try:
    import sift
except:
    print("FATAL : sift module not found !")


try:
    from sift import param
except:
    print("WARNING: Parameters file not found ! Taking default parameters")
    par = Enum(DoubleImSize=0,
    InitSigma=1.6,
    BorderDist=5,
    Scales=3,
    PeakThresh=255.0 * 0.04 / 3.0,
    EdgeThresh=0.06,
    EdgeThresh1=0.08,
    OriSigma=1.5,
    MatchRatio=0.73)
'''



class MainWin(QtGui.QMainWindow):
    
    def __init__(self):
        super(MainWin, self).__init__()
        self.step = 0
        self.folder = ""
        self.MAX_KP = 5000 #TODO: user-defined
        '''
        self.DoubleImSize
        self.InitSigma
        self.BorderDist
        self.Scales
        self.PeakThresh
        self.EdgeThresh
        self.EdgeThresh1
        self.OriSigma
        self.MatchRatio
        '''
        self.setDefaultPars()
        
        
        self.initUI()
        
    def initUI(self):      

        btn1 = QtGui.QPushButton("Select a directory", self)
        btn1.move(30, 50)
        btn2 = QtGui.QPushButton("Start alignment", self)
        btn2.move(120, 170)
        btn1.clicked.connect(self.buttonClicked)            
        btn2.clicked.connect(self.buttonClicked)
        btn1.adjustSize()
        btn2.adjustSize()
        
        self.pbar = QtGui.QProgressBar(self)
        self.pbar.setGeometry(30, 40, 300, 25)
        self.pbar.move(30,210)
        
        lbl0 = QtGui.QLabel(self)
        text0 = lbl0.setText("Please select a folder containing the images to align,\n and press \"start alignment\".")
        lbl0.move(30, 10)
        lbl0.adjustSize() 
        self.lbl1 = QtGui.QLabel(self)
        self.lbl1.setText("No folder selected yet")
        self.lbl1.move(35,85); self.lbl1.adjustSize()
        
        lbl2 = QtGui.QLabel(self)
        text2 = lbl2.setText("Run on device : ")
        lbl2.move(30, 110)
        self.combo = QtGui.QComboBox(self)
        self.combo.move(145, 110)
        self.detectDevices()


        
        
        
        
        
        
        

#        self.btn3 = QtGui.QPushButton("Edit SIFT parameters", self)
#        self.btn3.move(230, 50)
#        self.btn3.clicked.connect(self.buttonClicked)
#        self.btn3.adjustSize()

        self.text_DoubleImSize = QtGui.QLabel(self); self.text_DoubleImSize.setText("DoubleImSize :")
        self.text_DoubleImSize.move(400,20)
        self.edit_DoubleImSize=QtGui.QLineEdit(self); self.edit_DoubleImSize.setText(str(self.DoubleImSize))
        self.edit_DoubleImSize.move(500,25)
        self.edit_DoubleImSize.adjustSize()
#        self.edit_DoubleImSize.hide(); self.text_DoubleImSize.hide()
        
        self.text_InitSigma = QtGui.QLabel(self); self.text_InitSigma.setText("InitSigma :")
        self.text_InitSigma.move(400,50)
        self.edit_InitSigma=QtGui.QLineEdit(self); self.edit_InitSigma.setText(str(self.InitSigma))
        self.edit_InitSigma.move(500,55)
        self.edit_InitSigma.adjustSize()
#        self.edit_InitSigma.hide(); self.text_InitSigma.hide()
        
        self.text_BorderDist = QtGui.QLabel(self); self.text_BorderDist.setText("BorderDist :")
        self.text_BorderDist.move(400,80)
        self.edit_BorderDist=QtGui.QLineEdit(self); self.edit_BorderDist.setText(str(self.BorderDist))
        self.edit_BorderDist.move(500,85)
        self.edit_BorderDist.adjustSize()
#        self.edit_BorderDist.hide(); self.text_BorderDist.hide()
        
        self.text_Scales = QtGui.QLabel(self); self.text_Scales.setText("Scales :")
        self.text_Scales.move(400,110)
        self.edit_Scales=QtGui.QLineEdit(self); self.edit_Scales.setText(str(self.Scales))
        self.edit_Scales.move(500,115)
        self.edit_Scales.adjustSize()
#        self.edit_Scales.hide(); self.text_Scales.hide()
        
        self.text_PeakThresh = QtGui.QLabel(self); self.text_PeakThresh.setText("PeakThresh :")
        self.text_PeakThresh.move(400,140)
        self.edit_PeakThresh=QtGui.QLineEdit(self); self.edit_PeakThresh.setText(str(self.PeakThresh))
        self.edit_PeakThresh.move(500,145)
        self.edit_PeakThresh.adjustSize()
#        self.edit_PeakThresh.hide(); self.text_PeakThresh.hide()
        
        self.text_EdgeThresh = QtGui.QLabel(self); self.text_EdgeThresh.setText("EdgeThresh :")
        self.text_EdgeThresh.move(400,175)
        self.edit_EdgeThresh=QtGui.QLineEdit(self); self.edit_EdgeThresh.setText(str(self.EdgeThresh))
        self.edit_EdgeThresh.move(500,180)
        self.edit_EdgeThresh.adjustSize()
#        self.edit_EdgeThresh.hide(); self.text_EdgeThresh.hide()
        
        self.text_EdgeThresh1 = QtGui.QLabel(self); self.text_EdgeThresh1.setText("EdgeThresh1 :")
        self.text_EdgeThresh1.move(400,205)
        self.edit_EdgeThresh1=QtGui.QLineEdit(self); self.edit_EdgeThresh1.setText(str(self.EdgeThresh1))
        self.edit_EdgeThresh1.move(500,210)
        self.edit_EdgeThresh1.adjustSize()
#        self.edit_EdgeThresh1.hide(); self.text_EdgeThresh1.hide();
        
        self.text_OriSigma = QtGui.QLabel(self); self.text_OriSigma.setText("OriSigma :")
        self.text_OriSigma.move(400,235)
        self.edit_OriSigma=QtGui.QLineEdit(self); self.edit_OriSigma.setText(str(self.OriSigma))
        self.edit_OriSigma.move(500,240)
        self.edit_OriSigma.adjustSize()
#        self.edit_OriSigma.hide(); self.text_OriSigma.hide()
        
        self.text_MatchRatio = QtGui.QLabel(self); self.text_MatchRatio.setText("Match Ratio :")
        self.text_MatchRatio.move(400,265)
        self.edit_MatchRatio=QtGui.QLineEdit(self); self.edit_MatchRatio.setText(str(self.MatchRatio))
        self.edit_MatchRatio.move(500,270)
        self.edit_MatchRatio.adjustSize()
#        self.edit_MatchRatio.hide(); self.text_MatchRatio.hide()
        
        btn4 = QtGui.QPushButton("Reset default parameters", self)
        btn4.move(450, 295)
        btn4.clicked.connect(self.buttonClicked)
        btn4.adjustSize()
        
        
        
        self.statusBar()
        self.setGeometry(300, 300, 650, 350)
        self.setWindowTitle('SIFT main window')
        self.show()



        
    def buttonClicked(self):
        stext = self.sender().text()
        if stext == "Select a directory":
            dname = str(QtGui.QFileDialog.getExistingDirectory(self, "Select Directory"))
            self.lbl1.setText(dname)
            self.lbl1.adjustSize()
            self.folder = dname
        if stext == "Start alignment":
            self.step = self.step + 1
            self.pbar.setValue(self.step)
            self.runSift()
            
        '''
        if stext == "Edit SIFT parameters":
            self.edit_DoubleImSize.show(); self.text_DoubleImSize.show()
            self.edit_InitSigma.show(); self.text_InitSigma.show()
            self.edit_BorderDist.show(); self.text_BorderDist.show()
            self.edit_Scales.show(); self.text_Scales.show()
            self.edit_PeakThresh.show(); self.text_PeakThresh.show()
            self.edit_EdgeThresh.show(); self.text_EdgeThresh.show()
            self.edit_EdgeThresh1.show(); self.text_EdgeThresh1.show();
            self.edit_OriSigma.show(); self.text_OriSigma.show()
            self.edit_MatchRatio.show(); self.text_MatchRatio.show()
            self.btn3.setText("Hide SIFT parameters")
        if stext == "Hide SIFT parameters":  
            self.edit_DoubleImSize.hide(); self.text_DoubleImSize.hide()
            self.edit_InitSigma.hide(); self.text_InitSigma.hide()
            self.edit_BorderDist.hide(); self.text_BorderDist.hide()
            self.edit_Scales.hide(); self.text_Scales.hide()
            self.edit_PeakThresh.hide(); self.text_PeakThresh.hide()
            self.edit_EdgeThresh.hide(); self.text_EdgeThresh.hide()
            self.edit_EdgeThresh1.hide(); self.text_EdgeThresh1.hide();
            self.edit_OriSigma.hide(); self.text_OriSigma.hide()
            self.edit_MatchRatio.hide(); self.text_MatchRatio.hide()
            self.btn3.setText("Edit SIFT parameters")
        '''



    def runSift(self):
    
        if not(os.path.isdir(self.folder)):
            r = QtGui.QMessageBox.question(self, 'Error', "Please choose a folder first", QtGui.QMessageBox.Ok, QtGui.QMessageBox.Ok)
        else:
            rep = self.folder
            files = [ os.path.join(rep,f) for f in os.listdir(rep) if os.path.isfile(os.path.join(rep,f)) ]
            nbfiles = len(files)
            i=0
            dtype_kp = numpy.dtype([('x', numpy.float32),
                                    ('y', numpy.float32),
                                    ('scale', numpy.float32),
                                    ('angle', numpy.float32),
                                    ('desc', (numpy.uint8, 128))
                                    ])
#            keypoints_global = numpy.recarray(shape=(nbfiles,self.MAX_KP), dtype=dtype_kp)
#            match_global = numpy.recarray(shape=(nbfiles, self.MAX_KP, 2), dtype=dtype_kp)
            
            t0 = time.time()
            for filename in files[:]:
                '''
                Step 1 : keypoints computation
                '''
                self.statusBar().showMessage('[Step 1] Processing image ' + os.path.basename(filename))
                image = scipy.misc.imread(filename)
                plan = sift.SiftPlan(template=image, devicetype="gpu") #TODO: devicetype=..., device=device #user-defined
                kp = plan.keypoints(image) #TODO : pipeline of frames, instead of re-launching plan
                sh = kp.shape[0]
#                keypoints_global[i,sh] = kp
                if (i == 0): ref = copy.deepcopy(kp) #alignment is done on the first image #TODO
                i+=1
                self.statusBar().showMessage('[Step 1] Processing frame ' + os.path.basename(filename) + ' - ' + str(sh) + ' keypoints found')
                
                '''
                Step 2 : image matching
                '''
                if (i != 0): #alignment is done on the first image #TODO
                    mp = sift.MatchPlan() #TODO : pipeline of frames, instead of re-launching matchplan
                    m = mp.match(kp, ref)
                    self.statusBar().showMessage('[Step 2] Processing frame ' + os.path.basename(filename) + ' - ' + str(m.shape[0]) + ' match found')
                
                self.step = self.step + 100.0/nbfiles
                self.pbar.setValue(self.step)
            t1 = time.time()
            self.statusBar().showMessage('Process took ' + str(t1-t0)[0:4] + ' s')


    
    def setDefaultPars(self):
        self.DoubleImSize=0
        self.InitSigma=1.6
        self.BorderDist=5
        self.Scales=3
        self.PeakThresh=255.0 * 0.04 / 3.0
        self.EdgeThresh=0.06
        self.EdgeThresh1=0.08
        self.OriSigma=1.5
        self.MatchRatio=0.73


    def detectDevices(self):
    
        #TODO : import .opencl then list the devices
        self.combo.addItem("GPU")
        self.combo.activated[str].connect(self.onActivated) 
    
    
    def onActivated(self, text):
        #TODO : with @param text, select the platform and device
        self.statusBar().showMessage('Device selected : ' + text)
        
        
        








def main():
    
    app = QtGui.QApplication(sys.argv)
    ex = MainWin()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()
    
    
    
    




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for all preprocessing kernels.
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-06-20"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
import sift_pyocl as sift
from sift_pyocl.utils import calc_size

logger = getLogger(__file__)

if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

kernels = {"preprocess":8 , "gaussian":512}
for kernel in kernels:
    kernel_file = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), kernel + ".cl")
    kernel_src = open(kernel_file).read()
    program = pyopencl.Program(ctx, kernel_src).build("-D WORKGROUP=%s" % kernels[kernel])
    kernels[kernel] = program

def gaussian_cpu(sigma, size=None):
    """
    Calculate a 1D gaussian using numpy.
    This is the same as scipy.signal.gaussian

    @param sigma: width of the gaussian
    @param size: can be calculated as 1 + 2 * 4sigma
    """
    t0 = time.time()
    if not size:
        size = int(1 + 8 * sigma)
    x = numpy.arange(size) - (size - 1.0) / 2.0
    g = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
    g /= g.sum(dtype=numpy.float32)

    if PROFILE:
        logger.info("execution time: %.3fms on CPU" % (1e3 * (time.time() - t0)))
    return g

def gaussian_gpu_v1(sigma, size=None):
    """
    Calculate a 1D gaussian using pyopencl.
    This is the same as scipy.signal.gaussian

    @param sigma: width of the gaussian
    @param size: can be calculated as 1 + 2 * 4sigma
    """
    if not size:
        size = int(1 + 8 * sigma)
    g_gpu = pyopencl.array.empty(queue, size, dtype=numpy.float32, order="C")
    t0 = time.time()
    evt1 = kernels["preprocess"].gaussian(queue, (size,), (1,),
                                        g_gpu.data,  # __global     float     *data,
                                        numpy.float32(sigma),  # const        float     sigma,
                                        numpy.int32(size))  # const        int     SIZE
    sum_data = pyopencl.array.sum(g_gpu, dtype=numpy.float32, queue=queue)
    evt2 = kernels["preprocess"].divide_cst(queue, (size,), (1,),
                                          g_gpu.data,  # __global     float     *data,
                                          sum_data.data,  # const        float     sigma,
                                          numpy.int32(size))  # const        int     SIZE
    g = g_gpu.get()
    if PROFILE:
        logger.info("execution time: %.3fms; Kernel took %.3fms and %.3fms" % (1e3 * (time.time() - t0), 1e-6 * (evt1.profile.end - evt1.profile.start)
                , 1e-6 * (evt2.profile.end - evt2.profile.start)))

    return g

def gaussian_gpu_v2(sigma, size=None):
    """
    Calculate a 1D gaussian using pyopencl.
    This is the same as scipy.signal.gaussian.
    Only one kernel to

    @param sigma: width of the gaussian
    @param size: can be calculated as 1 + 2 * 4sigma
    """
    if not size:
        size = int(1 + 8 * sigma)
    g_gpu = pyopencl.array.empty(queue, size, dtype=numpy.float32, order="C")
    t0 = time.time()
    evt = kernels["gaussian"].gaussian(queue, (64,), (64,),
                                        g_gpu.data,  # __global     float     *data,
                                        numpy.float32(sigma),  # const        float     sigma,
                                        numpy.int32(size))  # const        int     SIZE
    g = g_gpu.get()
    if PROFILE:
        logger.info("execution time: %.3fms; Kernel took %.3fms" % (1e3 * (time.time() - t0), 1e-6 * (evt.profile.end - evt.profile.start)))
    return g


def show(ref, res, delta):
    pylab.ion()
    pylab.plot(ref, label="ref")
    pylab.plot(res, label="res")
    pylab.plot(delta, label="delta")
    pylab.legend()
    pylab.show()
    raw_input("enter")

class test_gaussian_v1(unittest.TestCase):

    def test_odd(self):
        """
        test odd kernel size
        """
        sigma = 3.0
        size = 27
        ref = gaussian_cpu(sigma, size)
        res = gaussian_gpu_v1(sigma, size)
        delta = ref - res
        if PROFILE:
            show (ref, res, delta)
        self.assert_(abs(ref - res).max() < 1e-6, "gaussian are the same ")

    def test_even(self):
        """
        test odd kernel size
        """
        sigma = 3.0
        size = 28
        ref = gaussian_cpu(sigma, size)
        res = gaussian_gpu_v1(sigma, size)
        delta = ref - res
        if PROFILE:
            show (ref, res, delta)
        self.assert_(abs(ref - res).max() < 1e-6, "gaussian are the same ")

class test_gaussian_v2(unittest.TestCase):

    def test_odd(self):
        """
        test odd kernel size
        """
        sigma = 3.0
        size = 27
        ref = gaussian_cpu(sigma, size)
        res = gaussian_gpu_v2(sigma, size)
        delta = ref - res
        if PROFILE:
            show (ref, res, delta)
        self.assert_(abs(ref - res).max() < 1e-6, "gaussian are the same ")

    def test_even(self):
        """
        test odd kernel size
        """
        sigma = 3.0
        size = 28
        ref = gaussian_cpu(sigma, size)
        res = gaussian_gpu_v2(sigma, size)
        delta = ref - res
        if PROFILE:
            show (ref, res, delta)
        self.assert_(abs(ref - res).max() < 1e-6, "gaussian are the same ")


def test_suite_gaussian():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_gaussian_v1("test_odd"))
    testSuite.addTest(test_gaussian_v1("test_even"))
    testSuite.addTest(test_gaussian_v2("test_odd"))
    testSuite.addTest(test_gaussian_v2("test_even"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_gaussian()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for image kernels
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc, scipy.ndimage, pylab
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
from test_image_functions import * #for Python implementation of tested functions
from test_image_setup import *
import sift_pyocl as sift
from sift_pyocl.utils import calc_size
logger = getLogger(__file__)
if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

SHOW_FIGURES = False
PRINT_KEYPOINTS = False


print "working on %s" % ctx.devices[0].name

'''
For Python implementation of tested functions, see "test_image_functions.py"
'''



class test_image(unittest.TestCase):
    def setUp(self):

        kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), "image.cl")
        kernel_src = open(kernel_path).read()
        self.program = pyopencl.Program(ctx, kernel_src).build()
        self.wg = (8, 1)



    def tearDown(self):
        self.mat = None
        self.program = None



    def test_gradient(self):
        """
        tests the gradient kernel (norm and orientation)
        """
        
        border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, scale, nb_keypoints, width, height, DOGS, g = local_maxmin_setup()
        self.mat = numpy.ascontiguousarray(g[1])
        self.height, self.width = numpy.int32(self.mat.shape)
        self.gpu_mat = pyopencl.array.to_device(queue, self.mat)
        self.gpu_grad = pyopencl.array.empty(queue, self.mat.shape, dtype=numpy.float32, order="C")
        self.gpu_ori = pyopencl.array.empty(queue, self.mat.shape, dtype=numpy.float32, order="C")
        self.shape = calc_size((self.width, self.height), self.wg)

        t0 = time.time()
        k1 = self.program.compute_gradient_orientation(queue, self.shape, self.wg, self.gpu_mat.data, self.gpu_grad.data, self.gpu_ori.data, self.width, self.height)
        res_norm = self.gpu_grad.get()
        res_ori = self.gpu_ori.get()
        t1 = time.time()
        ref_norm, ref_ori = my_gradient(self.mat)
        t2 = time.time()
        delta_norm = abs(ref_norm - res_norm).max()
        delta_ori = abs(ref_ori - res_ori).max()
        if (PRINT_KEYPOINTS):
            rmin, cmin = 0, 0
            rmax, cmax = rmin+6, cmin+6
            
            print res_norm[-rmax,cmin:cmax]
            print ""
            print ref_norm[-rmax,cmin:cmax]
            fig = pylab.figure()
            sp1 = fig.add_subplot(121)
            sp1.imshow(res_norm, interpolation="nearest")
            sp2 = fig.add_subplot(122)
            sp2.imshow(ref_norm, interpolation="nearest")
            fig.show()
            raw_input("enter")
        
        self.assert_(delta_norm < 1e-4, "delta_norm=%s" % (delta_norm))
        self.assert_(delta_ori < 1e-4, "delta_ori=%s" % (delta_ori))
        logger.info("delta_norm=%s" % delta_norm)
        logger.info("delta_ori=%s" % delta_ori)

        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Gradient computation took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))





    def test_local_maxmin(self):
        """
        tests the local maximum/minimum detection kernel
        """
        #local_maxmin_setup :
        border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, s, nb_keypoints, width, height, DOGS, g = local_maxmin_setup()
        self.s = numpy.int32(s) #1, 2, 3 ... not 4 nor 0.
        self.gpu_dogs = pyopencl.array.to_device(queue, DOGS)
        self.output = pyopencl.array.empty(queue, (nb_keypoints, 4), dtype=numpy.float32, order="C")
        self.output.fill(-1.0, queue) #memset for invalid keypoints
        self.counter = pyopencl.array.zeros(queue, (1,), dtype=numpy.int32, order="C")
        nb_keypoints = numpy.int32(nb_keypoints)
        self.shape = calc_size((DOGS.shape[1], DOGS.shape[0] * DOGS.shape[2]), self.wg) #it's a 3D vector !!

        t0 = time.time()
        k1 = self.program.local_maxmin(queue, self.shape, self.wg,
        	self.gpu_dogs.data, self.output.data,
       		border_dist, peakthresh, octsize, EdgeThresh0, EdgeThresh,
       		self.counter.data, nb_keypoints, self.s, width, height)

        res = self.output.get()
        self.keypoints1 = self.output #for further use
        self.actual_nb_keypoints = self.counter.get()[0] #for further use

        t1 = time.time()
        ref, actual_nb_keypoints2 = my_local_maxmin(DOGS, peakthresh, border_dist, octsize,
        	EdgeThresh0, EdgeThresh, nb_keypoints, self.s, width, height)
        t2 = time.time()

        #we have to sort the arrays, for peaks orders is unknown for GPU
        res_peaks = res[(res[:, 0].argsort(axis=0)), 0]
        ref_peaks = ref[(ref[:, 0].argsort(axis=0)), 0]
        res_r = res[(res[:, 1].argsort(axis=0)), 1]
        ref_r = ref[(ref[:, 1].argsort(axis=0)), 1]
        res_c = res[(res[:, 2].argsort(axis=0)), 2]
        ref_c = ref[(ref[:, 2].argsort(axis=0)), 2]
        #res_s = res[(res[:,3].argsort(axis=0)),3]
        #ref_s = ref[(ref[:,3].argsort(axis=0)),3]
        delta_peaks = abs(ref_peaks - res_peaks).max()
        delta_r = abs(ref_r - res_r).max()
        delta_c = abs(ref_c - res_c).max()

        if (PRINT_KEYPOINTS):
            print("keypoints after 2 steps of refinement: (s= %s, octsize=%s) %s" % (self.s, octsize, self.actual_nb_keypoints))
            #print("For ref: %s" %(ref_peaks[ref_peaks!=-1].shape))
            print res[0:self.actual_nb_keypoints]#[0:74]
            #print ref[0:32]

        self.assert_(delta_peaks < 1e-4, "delta_peaks=%s" % (delta_peaks))
        self.assert_(delta_r < 1e-4, "delta_r=%s" % (delta_r))
        self.assert_(delta_c < 1e-4, "delta_c=%s" % (delta_c))
        logger.info("delta_peaks=%s" % delta_peaks)
        logger.info("delta_r=%s" % delta_r)
        logger.info("delta_c=%s" % delta_c)


        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Local extrema search took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))





    def test_interpolation(self):
        """
        tests the keypoints interpolation kernel
        Requires the following: "self.keypoints1", "self.actual_nb_keypoints", 	"self.gpu_dog_prev", self.gpu_dog", 			"self.gpu_dog_next", "self.s", "self.width", "self.height", "self.peakthresh"
        """

        #interpolation_setup :
        border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, nb_keypoints, actual_nb_keypoints, width, height, DOGS, s, keypoints_prev, blur = interpolation_setup()

        # actual_nb_keypoints is the number of keypoints returned by "local_maxmin".
        #After the interpolation, it will be reduced, but we can still use it as a boundary.
        shape = calc_size(keypoints_prev.shape, self.wg)
        gpu_dogs = pyopencl.array.to_device(queue, DOGS)
        gpu_keypoints1 = pyopencl.array.to_device(queue, keypoints_prev)
        #actual_nb_keypoints = numpy.int32(len((keypoints_prev[:,0])[keypoints_prev[:,1] != -1]))
        start_keypoints = numpy.int32(0)
        actual_nb_keypoints = numpy.int32(actual_nb_keypoints)
        InitSigma = numpy.float32(1.6) #warning: it must be the same in my_keypoints_interpolation
        t0 = time.time()
        k1 = self.program.interp_keypoint(queue, shape, self.wg,
        	gpu_dogs.data, gpu_keypoints1.data, start_keypoints, actual_nb_keypoints,
        	peakthresh, InitSigma, width, height)
        res = gpu_keypoints1.get()

        t1 = time.time()
        ref = numpy.copy(keypoints_prev) #important here
        for i, k in enumerate(ref[:nb_keypoints, :]):
            ref[i] = my_interp_keypoint(DOGS, s, k[1], k[2], 5, peakthresh, width, height)

        t2 = time.time()


        #we have to compare keypoints different from (-1,-1,-1,-1)
        res2 = res[res[:, 1] != -1]
        ref2 = ref[ref[:, 1] != -1]


        if (PRINT_KEYPOINTS):
            print("[s=%s]Keypoints before interpolation: %s" % (s, actual_nb_keypoints))
            #print keypoints_prev[0:10,:]
            print("[s=%s]Keypoints after interpolation : %s" % (s, res2.shape[0]))
            print res[0:actual_nb_keypoints]#[0:10,:]
            #print("Ref:")
            #print ref[0:32,:]


        delta = abs(ref2 - res2).max()
        self.assert_(delta < 1e-4, "delta=%s" % (delta))
        logger.info("delta=%s" % delta)

        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Keypoints interpolation took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))







def test_suite_image():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_image("test_gradient"))
    testSuite.addTest(test_image("test_local_maxmin"))
    testSuite.addTest(test_image("test_interpolation"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_image()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)

